import streamlit as st

# Page paths relative to entry point (run_app.py)
welcome_page = st.Page("app/pages/welcome.py", title="🥚 Welcome", default=True)
jobs_page = st.Page("app/pages/jobs/jobs.py", title="⏳ Jobs")
designs_page = st.Page("app/pages/designs/designs.py", title="🐣 Designs")

rf_scaffold_design_page = st.Page("app/pages/rfdiffusion/scaffold_design.py", title="🏗️ Scaffold design")
rf_binder_design_page = st.Page("app/pages/rfdiffusion/binder_design.py", title="🧬 Binder design")
rf_binder_diversification_page = st.Page(
    "app/pages/rfdiffusion/binder_diversification.py", title="♻️ Binder diversification"
)

bindcraft_binder_design_page = st.Page("app/pages/bindcraft/bindcraft_binder_design.py", title="⚒️ Binder design")

admin_debug_page = st.Page("app/pages/debug.py", title="🖥️ Debug")
admin_import_export_page = st.Page("app/pages/import_export.py", title="📦 Import & Export")



from datetime import datetime
from time import time
import boto3
from botocore.exceptions import ClientError

from mypy_boto3_s3.client import S3Client
from mypy_boto3_omics import OmicsClient


class AWSSessionManager:
    def __init__(self, assume_role: str = None, region_name="us-east-1"):
        self.assume_role = assume_role
        self.region_name = region_name
        self._AWS_CLIENTS = {}
        self._AWS_CLIENTS_EXPIRATION = {}

    @property
    def s3(self) -> S3Client:
        return self.get_client("s3")

    @property
    def omics(self) -> OmicsClient:
        return self.get_client("omics")

    def get_client(self, service, refresh_minutes_before_expiration=30):
        expired = (
            service in self._AWS_CLIENTS_EXPIRATION
            and self._AWS_CLIENTS_EXPIRATION[service].timestamp() - time() < 60 * refresh_minutes_before_expiration
        )
        if service not in self._AWS_CLIENTS or expired:
            session, expiration_datetime = self._create_session()
            self._AWS_CLIENTS[service] = session.client(service)
            if expiration_datetime is not None:
                self._AWS_CLIENTS_EXPIRATION[service] = expiration_datetime
        return self._AWS_CLIENTS[service]

    def _create_session(self) -> tuple[boto3.Session, datetime | None]:
        sts_client = boto3.client("sts")

        if self.assume_role is not None:
            assumed_role_object = sts_client.assume_role(RoleArn=self.assume_role, RoleSessionName="OVO")

            # The response contains temporary credentials to the assumed role.
            credentials = assumed_role_object["Credentials"]

            session = boto3.Session(
                aws_access_key_id=credentials["AccessKeyId"],
                aws_secret_access_key=credentials["SecretAccessKey"],
                aws_session_token=credentials["SessionToken"],
                region_name=self.region_name,
            )
            expiration_datetime = credentials["Expiration"]
        else:
            session = boto3.Session()
            expiration_datetime = None

        return session, expiration_datetime

    def get_account_id(self):
        # Use the STS client to get account information
        sts_client = self.get_client("sts")
        response = sts_client.get_caller_identity()
        return response["Account"]

    def get_latest_workflow_id(self, workflow_name: str):
        limit = 100
        items = self.omics.list_workflows(type="PRIVATE", name=workflow_name, maxResults=limit)["items"]
        if len(items) >= limit:
            # We could add pagination to solve this, at least make sure we raise an error for now
            raise ValueError(f"Too many workflows with name '{workflow_name}' found in AWS (>{limit})")
        if not items:
            raise ValueError(f"HealthOmics workflow with name '{workflow_name}' was not found in AWS")
        latest_workflow = sorted(items, key=lambda x: x["creationTime"], reverse=True)[0]
        return latest_workflow["id"]

    def s3_file_exists(self, s3_uri: str):
        assert s3_uri.startswith("s3://"), f"Invalid S3 URI: {s3_uri}"
        bucket, key = s3_uri[5:].split("/", maxsplit=1)
        try:
            self.s3.head_object(Bucket=bucket, Key=key)
            return True
        except ClientError as e:
            # Check if the error is "Not Found"
            if e.response["Error"]["Code"] in ("NoSuchKey", "404", 404):
                return False
            else:
                print("S3 ERROR", e, e.response)
                raise e

    def get_s3_bytes(self, s3_uri: str):
        assert s3_uri.startswith("s3://"), f"Invalid S3 URI: {s3_uri}"
        bucket, key = s3_uri[5:].split("/", maxsplit=1)
        response = self.s3.get_object(Bucket=bucket, Key=key)
        return response["Body"].read()



import dataclasses
import json
import os
from dataclasses import dataclass, field
from io import StringIO
from pathlib import Path

import yaml
from pydantic import model_validator
from pydantic_settings import BaseSettings, SettingsConfigDict

from ovo.cli.common import console, OVOCliError, OVONotInitializedError


class BaseConfigModel(BaseSettings):
    model_config = SettingsConfigDict(
        # Enable automatic type conversion for all child models
        from_attributes=True,
        # Raise error on fields that do not match the model
        extra="forbid",
        # Prefix for env vars - override any config value with OVO_SOME_FIELD_SUBFIELD=123
        env_prefix="OVO_",
    )


@dataclass
class DBConfig:
    url: str | None
    verbose: bool = False


@dataclass
class AWSConfig:
    region: str
    assume_role_arn: str


@dataclass
class StorageConfig:
    path: str | None
    verbose: bool = False
    aws: AWSConfig | None = field(default=None)


@dataclass
class SchedulerConfig:
    type: str
    name: str
    workdir: str
    aws: AWSConfig | None = field(default=None)
    submission_args: dict = field(default_factory=dict)


@dataclass
class AuthConfig:
    # Users who can access Debug page and run commands on the server through the web UI
    admin_users: list[str] = field(default_factory=list)
    # Enable native Streamlit authentication
    streamlit_auth: bool = field(default=False)
    # Allow users to visit private projects if they have the link (containing the project UID)
    allow_private_project_link_access: bool = field(default=True)
    # Hide admin access warning if admin_users is set but no auth is configured
    hide_admin_warning: bool = field(default=False)
    # Always generate a required token to access the app
    always_require_token: bool = field(default=False)


@dataclass
class ConfigProps:
    pyrosetta_license: bool = False
    read_only: bool = False
    rfdiffusion_backbones_limit: int = 1000
    rfdiffusion_backbones_limit_admin: int = 5000
    mpnn_sequences_limit: int = 100


@dataclass
class TemplatesConfig:
    welcome: str | None = None
    welcome_appendix: str | None = None


class OVOConfig(BaseConfigModel):
    # Absolute path to config directory (not stored in YAML, initialized in load_config)
    dir: str = field()
    # Absolute path to reference files directory (model weights, etc)
    reference_files_dir: str = field()
    # Database instance that manages all interactions with the DB
    db: DBConfig = field(default_factory=DBConfig)
    # Authentication config
    auth: AuthConfig = field(default_factory=AuthConfig)
    # Storage instance that manages PDBs and other files
    storage: StorageConfig = field(default_factory=StorageConfig)
    # Key of the default scheduler to use for job submissions
    default_scheduler: str | None = None
    # Key of the local scheduler to use for local job submissions like RFdiffusion preview
    local_scheduler: str | None = None
    # Dictionary of scheduler instances (scheduler key -> instance) that enable scheduling jobs
    schedulers: dict[str, SchedulerConfig] = field(default_factory=dict)
    # Nextflow home directory
    nextflow_home: str = field(default="./nextflow")
    # Text content configuration
    templates: TemplatesConfig = field(default_factory=TemplatesConfig)
    # Remaining configuration settings
    props: ConfigProps = field(default_factory=ConfigProps)
    # Freeform plugin configuration (plugin module name -> config dict)
    plugins: dict[str, dict] = field(default_factory=dict)

    @model_validator(mode="after")
    def resolve_relative_paths(self):
        """Resolve relative paths to absolute paths based on config directory

        Config paths are stored as relative to config directory,
        but in the config object instance they are resolved to full absolute paths.
        """

        def make_absolute_if_relative(path: str) -> str:
            """Convert relative path to absolute by prepending home_dir"""
            if path and not os.path.isabs(path) and "://" not in path:
                return os.path.normpath(os.path.join(self.dir, path))
            return path

        # Resolve DB path
        if self.db.url.startswith("sqlite:///"):
            self.db.url = "sqlite:///" + make_absolute_if_relative(self.db.url.removeprefix("sqlite:///"))

        # Resolve absolute paths
        self.reference_files_dir = make_absolute_if_relative(self.reference_files_dir)
        self.storage.path = make_absolute_if_relative(self.storage.path)
        self.nextflow_home = make_absolute_if_relative(self.nextflow_home)
        for scheduler_name, scheduler_config in self.schedulers.items():
            scheduler_config.workdir = make_absolute_if_relative(scheduler_config.workdir)
            for path_arg in ["config"]:
                if path_arg in scheduler_config.submission_args:
                    scheduler_config.submission_args[path_arg] = make_absolute_if_relative(
                        scheduler_config.submission_args[path_arg]
                    )

        for templ_field in dataclasses.fields(self.templates):
            path = getattr(self.templates, templ_field.name)
            setattr(self.templates, templ_field.name, make_absolute_if_relative(path))

        return self

    @classmethod
    def default(cls, props: ConfigProps, default_profile=None, admin_users: list[str] = None) -> str:
        if default_profile is None:
            default_profile = "conda"
        props_rows = "\n".join(f"  {k}: {json.dumps(v)}" for k, v in props.__dict__.items())
        default_scheduler_key = f"local_{default_profile}"
        return f"""
db:
  url: sqlite:///ovo.db
  verbose: false
reference_files_dir: ./reference_files
nextflow_home: ./nextflow
auth:
  admin_users: {json.dumps(admin_users or [])}
  allow_private_project_link_access: true
storage:
  verbose: false
  path: ./storage
# Example of a storage path on AWS S3:
#  path: s3://bucket-name/ovo/storage/
#  aws:
#    region: us-east-1
#    assume_role_arn: arn:aws:iam::YOUR_ACCOUNT_ID:role/YOUR-ROLE (or null if not assuming a different role)
props:
{props_rows}
default_scheduler: {default_scheduler_key}
local_scheduler: {default_scheduler_key}
schedulers:
  {default_scheduler_key}:
    name: Local with {default_profile}
    type: NextflowScheduler
    submission_args:
      # using CPU version of dependencies, remove 'cpu_env' if your local machine has a GPU
      profile: {default_profile},cpu_env
      max_memory: 8GB
      config: ./nextflow_local.config
    workdir: ./workdir
# Plugin-specific configuration can go here, for example:
# plugins:
#   ovo_plugin_name:
#     key: value
# Uncomment to customize web page content
# templates:
#     welcome: templates/welcome.md
#     welcome_appendix: templates/welcome_appendix.md
"""

    @staticmethod
    def default_nextflow_config():
        return """
process {
    // process settings
    
    executor = "local"  // Executor, for example "pbs", "slurm", ...
    
    beforeScript = ""   // Script to execute before each job, for example "module add singularity"
    queue = null        // Job queue, for example "long"
    
    // customize resources per process
    //withLabel: "colabdesign|rfdiffusion" {
    //    queue = 'gpu-t4'
    //    memory = '16 GB'
    //    cpus = 4
    //    // alternatively to memory and cpus, you can specify cluster job submission options directly
    //    // clusterOptions = "-l select=1:ncpus=4:mem=20gb"
    //}
}
//
//params {
//    // override default parameters here
//    foo = "bar"
//}
"""


DEFAULT_OVO_HOME = os.path.abspath(os.path.expanduser("~/ovo/"))
GLOBAL_OVO_HOME = None
site_packages_dir = os.path.dirname(os.path.dirname(os.path.dirname(__file__)))
global_config_flag = os.path.join(site_packages_dir, "ovo.global_config")
if os.path.exists(global_config_flag):
    with open(global_config_flag) as f:
        GLOBAL_OVO_HOME = f.read().strip() or None

# Self-consistency check that default values are valid
default_data = yaml.safe_load(StringIO(OVOConfig.default(props=ConfigProps())))
default_data["dir"] = DEFAULT_OVO_HOME
OVOConfig.model_validate(default_data)


def save_global_home_dir(home_dir: str):
    # Ensure the directory is writable before attempting to write
    home_dir_path = os.path.dirname(global_config_flag)
    if not os.access(home_dir_path, os.W_OK):
        raise OVOCliError(f"Cannot write to the package directory: {home_dir_path}. Insufficient permissions.")
    try:
        with open(global_config_flag, "wt") as f:
            f.write(os.path.abspath(home_dir) + "\n")
    except OSError as e:
        raise OVOCliError(f"Failed to write global config flag at {global_config_flag}: {e}")


def get_shell_config_path():
    """Get path to .bashrc or .zshrc based on the current shell."""
    shell = os.path.basename(os.environ.get("SHELL", ""))
    home = str(Path.home())
    config_path = None
    if "zsh" in shell:
        config_path = os.path.join(home, ".zshrc")
    elif "bash" in shell:
        config_path = os.path.join(home, ".bashrc")
    if config_path and os.path.exists(config_path):
        return config_path
    return None


def get_source_command():
    """Print the command for the user to source the shell config file."""
    shell_config_path = get_shell_config_path()
    command = f"\n[bold]source {shell_config_path}[/bold] # load env variables"
    if active_env := os.getenv("CONDA_DEFAULT_ENV"):
        command += f"\n[bold]conda activate {active_env}[/bold] # re-activate this environment"
    return command


def load_config(home_dir: str) -> OVOConfig:
    """Load OVOConfig from config dir config.yml"""

    config_path = os.path.join(home_dir, "config.yml")

    if not os.path.exists(home_dir) or not os.path.exists(config_path):
        # Config dir does not exist yet, print error and exit
        if home_dir != DEFAULT_OVO_HOME:
            # Custom OVO_HOME set but invalid
            if not os.path.exists(home_dir):
                raise OVONotInitializedError(f"OVO_HOME directory set but is not accessible: {home_dir}")
            else:
                raise OVONotInitializedError(f"OVO config file not found in OVO_HOME: {home_dir}")
        else:
            # Default OVO_HOME, raise not initialized error
            raise OVONotInitializedError("OVO not initialized")

    console.print(f"[bold]OVO home[/bold] [green]{home_dir}[/green]")

    with open(config_path, "r") as f:
        data = yaml.safe_load(f)
        data["dir"] = os.path.abspath(home_dir)

    config = OVOConfig(**data)

    if config.props.read_only:
        console.print("[blue][bold]Running in read-only mode[/bold] (configured in config.yml props.read_only)[/blue]")

    return config


def save_default_config(
    home_dir, config_props: ConfigProps, default_profile: str = None, admin_users: list[str] = None
) -> str:
    config_path = os.path.join(home_dir, "config.yml")

    with open(config_path, "w") as f:
        f.write(OVOConfig.default(props=config_props, default_profile=default_profile, admin_users=admin_users))

    with open(os.path.join(home_dir, "nextflow_local.config"), "w") as f:
        f.write(OVOConfig.default_nextflow_config())

    return config_path



import typer
import os
from rich.prompt import Prompt, Confirm
from ovo.cli.common import console, OVOCliError, download_files, init_nextflow
from ovo.core.configuration import (
    ConfigProps,
    DEFAULT_OVO_HOME,
    get_source_command,
    get_shell_config_path,
    save_default_config,
)
import shutil

app = typer.Typer(pretty_exceptions_enable=False, help="OVO initialization commands")

RFDIFFUSION_MODEL_FILES = [
    # URL, local path, hash (SHA256: calculate with "shasum -a 256 my_file" or "sha256sum my_file")
    (
        "https://files.ipd.uw.edu/pub/RFdiffusion/6f5902ac237024bdd0c176cb93063dc4/Base_ckpt.pt",
        "rfdiffusion_models/Base_ckpt.pt",
        "0fcf7d7c32b4848030aca3a051e6768de194616f96ba6c38186351a33bfc6eca",
    ),
    (
        "https://files.ipd.uw.edu/pub/RFdiffusion/e29311f6f1bf1af907f9ef9f44b8328b/Complex_base_ckpt.pt",
        "rfdiffusion_models/Complex_base_ckpt.pt",
        "76e4e260aefee3b582bd76b77ab95d2592e64f00c51bf344968ab9239f3250bc",
    ),
    (
        "https://files.ipd.uw.edu/pub/RFdiffusion/f572d396fae9206628714fb2ce00f72e/Complex_beta_ckpt.pt",
        "rfdiffusion_models/Complex_beta_ckpt.pt",
        "5a0b1cafc23c60b1aabcec1e49391986ac4fd02cc1b6b4cc41714ca9fe882e9e",
    ),
    (
        "https://files.ipd.uw.edu/pub/RFdiffusion/5532d2e1f3a4738decd58b19d633b3c3/ActiveSite_ckpt.pt",
        "rfdiffusion_models/ActiveSite_ckpt.pt",
        "beca1f672049161df0bc6a2d2523828f19fd9c8a2b449988e246dde42e7ea986",
    ),
    (
        "https://files.ipd.uw.edu/pub/RFdiffusion/74f51cfb8b440f50d70878e05361d8f0/InpaintSeq_ckpt.pt",
        "rfdiffusion_models/InpaintSeq_ckpt.pt",
        "3b71b2b954e87d46b75a88ba64e0420fbf27f592604b10b6c3561b8c8ab70ab6",
    ),
]

ALPHAFOLD_MODEL_FILES = [
    (
        "https://storage.googleapis.com/alphafold/alphafold_params_2022-12-06.tar",
        "alphafold_models",
        "36d4b0220f3c735f3296d301152b738c9776d16981d054845a68a1370b26cfe3",
    )
]

ESM1V_MODEL_FILES = [
    (
        "https://dl.fbaipublicfiles.com/fair-esm/models/esm1v_t33_650M_UR90S_1.pt",
        "esm_models/esm1v_t33_650M_UR90S_1.pt",
        "9519ee60f1cddad3c101afb1f42612499e188534969c3f682e94850870f70433",
    ),
    (
        "https://dl.fbaipublicfiles.com/fair-esm/models/esm1v_t33_650M_UR90S_2.pt",
        "esm_models/esm1v_t33_650M_UR90S_2.pt",
        "5b7b095e8eafc53ccfe5994b954fb756bfe7a081f22b4caa1ed59b77b90bcf81",
    ),
    (
        "https://dl.fbaipublicfiles.com/fair-esm/models/esm1v_t33_650M_UR90S_3.pt",
        "esm_models/esm1v_t33_650M_UR90S_3.pt",
        "bc5cb2f2a1b35def284e2b2833ae58c803ca9c61b16b72c1dd54c54e76df0b67",
    ),
    (
        "https://dl.fbaipublicfiles.com/fair-esm/models/esm1v_t33_650M_UR90S_4.pt",
        "esm_models/esm1v_t33_650M_UR90S_4.pt",
        "44750a28c09f7ba9e7ccb7aeaba812cbbe90eb2a8a2c658dc5fa165f7090a15a",
    ),
    (
        "https://dl.fbaipublicfiles.com/fair-esm/models/esm1v_t33_650M_UR90S_5.pt",
        "esm_models/esm1v_t33_650M_UR90S_5.pt",
        "69ffd06be29aaaf105eda919f23e5ac4a6872e7907fbf1f087fd942abdb3adf7",
    ),
]

ESM_IF_MODEL_FILES = [
    (
        "https://dl.fbaipublicfiles.com/fair-esm/models/esm_if1_gvp4_t16_142M_UR50.pt",
        "esm_models/esm_if1_gvp4_t16_142M_UR50.pt",
        "be4ba36edec22a9bfaa4946ff6b2815f1f19d8a3d7e0eada8b796d5a0eae9fd4",
    )
]

BOLTZ_MODEL_FILES = {
    (
        "https://huggingface.co/boltz-community/boltz-1/resolve/main/boltz1_conf.ckpt",
        "boltz_models/boltz1_conf.ckpt",
        "fea245d912c570ec117b2277c2719f312a6fc109c07b6f6ef741690ee775c2f5",
    ),
    (
        "https://huggingface.co/boltz-community/boltz-1/resolve/main/ccd.pkl",
        "boltz_models/ccd.pkl",
        "2d3b2f03a3c5665944adba51e33263511e51b21c9cd05d902f9c4b7c1e58d2f4",
    ),
}


@app.command()
def home(
    home_dir: str | None = typer.Argument(None, help="OVO home directory"),
    yes: bool = typer.Option(False, "--yes", "-y", help="Confirm all without prompting"),
    no_env: bool = typer.Option(False, "--no-env", help="Do not set OVO_HOME in shell .bashrc/.zshrc file"),
    default_profile: str = typer.Option(None, "--profile", help="Default nextflow profile to use in default scheduler"),
):
    """Initialize the OVO home directory"""

    console.print("[bold]OVO initialization[/bold]")

    if home_dir is None:
        console.print("""
[bold]Create OVO home directory[/bold]

This directory will contain all OVO files:

- [bold]config.yml[/bold][gray]______[/gray]configuration file
- [bold]ovo.db[/bold][gray]__________[/gray]SQLite database (stored as a single file)
- [bold]workdir[/bold][gray]_________[/gray]Working directory for Nextflow workflows
- [bold]storage[/bold][gray]_________[/gray]PDB design files and other permanent results
- [bold]reference_files[/bold][gray]_[/gray]Downloaded model weights and other reference files

All paths can be customized later in the [bold]config.yml[/bold] file.
""")
        home_dir = yes or Prompt.ask(
            prompt="Enter path, or press [bold]Enter[/bold] to select the default",
            default=os.getenv("OVO_HOME", DEFAULT_OVO_HOME),  # Use OVO_HOME env var if set already
        )

    home_dir = os.path.abspath(os.path.expanduser(home_dir))

    while os.path.exists(home_dir):
        if yes:
            raise FileExistsError(f"Home directory already exists: {home_dir}")
        # show error f"Home directory already exists: {home_dir}"
        console.print(f"[red]Error:[/red] Home directory already exists: [bold]{home_dir}[/bold]")
        home_dir = Prompt.ask("Please choose a different path for the new OVO Home directory")
        home_dir = os.path.abspath(os.path.expanduser(home_dir))

    if not yes and home_dir != DEFAULT_OVO_HOME:
        if not Confirm.ask(f"Confirm directory: [bold]{home_dir}[/bold]"):
            console.print("Aborted")
            raise typer.Exit()

    if not default_profile:
        default_profile = "conda"
        if not yes:
            console.print("""
    OVO uses Nextflow to run workflows and manage software environments.
    How do you want the default OVO scheduler to manage dependencies?""")
            default_profile = Prompt.ask(
                "\nSelect profile, or press [bold]Enter[/bold] to select the default",
                choices=["conda", "singularity", "apptainer", "docker", "podman"],
                default=default_profile,
            )

    # ask about using conda when docker is not on PATH
    if not shutil.which(default_profile):
        console.print(
            f"[yellow]WARNING:[/yellow] {default_profile} not found on PATH, please make sure to install or activate {default_profile} or change the default scheduler profile (docker, singularity, apptainer, ...)"
        )

    config_props = ConfigProps()
    config_props.pyrosetta_license = yes or Confirm.ask(
        "\nEnable fastrelax? In case of commercial use, this requires a PyRosetta license"
    )

    admin_users = []
    if os.environ.get("USER"):
        # we don't want to flood the user with too many questions, enable by default
        admin_users.append(os.environ["USER"])

    os.makedirs(home_dir, exist_ok=True)

    config_path = save_default_config(
        home_dir=home_dir,
        config_props=config_props,
        default_profile=default_profile,
        admin_users=admin_users,
    )

    console.print(f"\n[green]✔[/green] Initialized OVO config: [bold]{config_path}[/bold]")

    if home_dir != DEFAULT_OVO_HOME and not no_env:
        console.print("\nYou will need to set the OVO_HOME environment variable to use this home directory.")
        export_command = f'export OVO_HOME="{home_dir}"'

        if (shell_config_path := get_shell_config_path()) and (
            yes or Confirm.ask(f"\nAdd the OVO_HOME dir to {shell_config_path}?")
        ):
            with open(shell_config_path, "a") as f:
                f.write(f"\n{export_command}")
            console.print(f"\n✔ Added to {shell_config_path}. Restart your terminal or run:")
            console.print(get_source_command())
        else:
            console.print("\nPlease set this environment variable manually:")
            console.print(f"[bold green]{export_command}[/bold green]")

    console.print("\nNext step: Initialize the preview workflow using [bold]ovo init preview[/bold]")


@app.command()
def preview():
    """Initialize nextflow, reference files and environment for the RFdiffusion preview workflow"""
    from ovo import config, local_scheduler

    # Initialize nextflow
    console.print("[bold]Initializing nextflow...[/bold]")
    init_nextflow()
    console.print("[bold][green]✔[/green] Nextflow initialized successfully[/bold]")

    # Download RFdiffusion base model weights
    download_files(destination_dir=config.reference_files_dir, file_list=RFDIFFUSION_MODEL_FILES)

    # Submit RFdiffusion preview (which also initializes the RFdiffusion conda environment)
    ovo_resources_path = os.path.abspath(os.path.join(os.path.dirname(__file__), "..", "resources"))
    process, job_id = local_scheduler.submit(
        "rfdiffusion-backbone",
        params={
            "contig": "A82-87/10/A92-97",
            "input_pdb": os.path.join(ovo_resources_path, "examples/inputs/5ELI_A.pdb"),
            "num_designs": 1,
            "run_parameters": "diffuser.T=2",
        },
        submission_args=dict(sync=True),
    )
    process.wait()
    if process.returncode == 0:
        console.print("[bold][green]✔[/green] RFdiffusion preview completed successfully[/bold]")
        console.print("Example unconditional design output saved to:")
        console.print(local_scheduler.get_output_dir(job_id))
    else:
        console.print("RFdiffusion workflow [red]FAILED[/red], please see errors above")
        exit(process.returncode)

    console.print("\nNext step: Set up end-to-end RFdiffusion workflow using [bold]ovo init rfdiffusion[/bold]")
    console.print("           Or already explore OVO web app using [bold]ovo app[/bold]")


@app.command()
def rfdiffusion(scheduler: str = None):
    """Initialize nextflow, reference files and environment for the RFdiffusion end-to-end workflow"""
    from ovo import config, get_scheduler

    # Initialize nextflow
    init_nextflow()

    # Download RFdiffusion base model weights
    download_files(destination_dir=config.reference_files_dir, file_list=RFDIFFUSION_MODEL_FILES)

    # Download AlphaFold weights
    download_files(destination_dir=config.reference_files_dir, file_list=ALPHAFOLD_MODEL_FILES)

    scheduler = get_scheduler(scheduler_key=scheduler or config.default_scheduler)

    # Submit RFdiffusion scaffold end-to-end workflows
    ovo_resources_path = os.path.abspath(os.path.join(os.path.dirname(__file__), "..", "resources"))
    process, job_id = scheduler.submit(
        "rfdiffusion-end-to-end",
        params={
            "rfdiffusion_contig": "A82-87/10/A92-97",
            "rfdiffusion_input_pdb": os.path.join(ovo_resources_path, "examples/inputs/5ELI_A.pdb"),
            "rfdiffusion_num_designs": 1,
            "rfdiffusion_run_parameters": "diffuser.T=2",
            "mpnn_num_sequences": 2,
            "design_type": "scaffold",
            "refolding_tests": "af2_model_1_ptm_ft_1rec",
        },
        submission_args=dict(sync=True),
    )
    process.wait()
    if process.returncode == 0:
        console.print("[bold][green]✔[/green] RFdiffusion scaffold preview completed successfully[/bold]")
        console.print("Example scaffold design output saved to:")
        console.print(scheduler.get_output_dir(job_id))
    else:
        console.print("RFdiffusion workflow [red]FAILED[/red], please see errors above")
        exit(process.returncode)

    # Submit RFdiffusion binder end-to-end workflow with LigandMPNN
    process, job_id = scheduler.submit(
        "rfdiffusion-end-to-end",
        params={
            "rfdiffusion_contig": "A20-130/0 10",
            "rfdiffusion_input_pdb": os.path.join(ovo_resources_path, "examples/inputs/5ELI_A.pdb"),
            "rfdiffusion_num_designs": 1,
            "rfdiffusion_run_parameters": "diffuser.T=15",
            "mpnn_num_sequences": 2,
            "mpnn_fastrelax_cycles": 0,
            "disable_pyrosetta_scoring": not config.props.pyrosetta_license,
            "design_type": "binder",
            "refolding_tests": "af2_model_1_multimer_tt_3rec",
        },
        submission_args=dict(sync=True),
    )
    process.wait()
    if process.returncode == 0:
        console.print("[bold][green]✔[/green] RFdiffusion binder preview completed successfully[/bold]")
        console.print("Example binder design output saved to:")
        console.print(scheduler.get_output_dir(job_id))
    else:
        console.print("RFdiffusion workflow [red]FAILED[/red], please see errors above")
        exit(process.returncode)

    if config.props.pyrosetta_license:
        # Submit RFdiffusion binder end-to-end workflow with FastRelax
        process, job_id = scheduler.submit(
            "rfdiffusion-end-to-end",
            params={
                "rfdiffusion_contig": "A20-130/0 10",
                "rfdiffusion_input_pdb": os.path.join(ovo_resources_path, "examples/inputs/5ELI_A.pdb"),
                "rfdiffusion_num_designs": 1,
                "rfdiffusion_run_parameters": "diffuser.T=15",
                "mpnn_fastrelax_cycles": 1,
                "design_type": "binder",
                "refolding_tests": "af2_model_1_multimer_tt_3rec",
            },
            submission_args=dict(sync=True),
        )
        process.wait()
        if process.returncode == 0:
            console.print("[bold][green]✔[/green] RFdiffusion binder preview completed successfully[/bold]")
            console.print("Example binder design output saved to:")
            console.print(scheduler.get_output_dir(job_id))
        else:
            console.print("RFdiffusion workflow [red]FAILED[/red], please see errors above")
            exit(process.returncode)

    console.print("\nNext step: Run OVO web app using [bold]ovo app[/bold]")


@app.command()
def proteinqc(
    tool_keys: str = typer.Option(
        "all",
        "--tools",
        help="Comma-separated list of ProteinQC tools to use, or 'all' to use all available tools (supported by the scheduler)",
    ),
    scheduler: str = None,
):
    """Initialize nextflow, reference files and environment for the ProteinQC workflow"""
    from ovo import config, get_scheduler
    from ovo.core.database.models_proteinqc import PROTEINQC_TOOLS, ESM_1V, ESM_IF
    from ovo.core.logic.proteinqc_logic import get_available_tools

    # Initialize nextflow
    init_nextflow()

    scheduler = get_scheduler(scheduler_key=scheduler or config.default_scheduler)

    if tool_keys != "all":
        tool_keys = tool_keys.split(",")
        all_tools_by_key = {tool.tool_key: tool for tool in PROTEINQC_TOOLS}
        tools = []
        for key in tool_keys:
            if key not in all_tools_by_key:
                raise OVOCliError(
                    f"Invalid ProteinQC tool '{key}', available tools: {','.join(all_tools_by_key.keys())}"
                )
            tools.append(all_tools_by_key[key])
    else:
        tools = get_available_tools(PROTEINQC_TOOLS, scheduler)
        tool_keys = [tool.tool_key for tool in tools]

    # Download ESM1v weights
    if ESM_1V in tools:
        download_files(destination_dir=config.reference_files_dir, file_list=ESM1V_MODEL_FILES)

    # Download ESM-IF weights
    if ESM_IF in tools:
        download_files(destination_dir=config.reference_files_dir, file_list=ESM_IF_MODEL_FILES)

    # Submit ProteinQC workflow
    ovo_resources_path = os.path.abspath(os.path.join(os.path.dirname(__file__), "..", "resources"))
    input_path = os.path.join(ovo_resources_path, "examples/inputs/5ELI_A.pdb")

    chains = ["A"]
    process, job_id = scheduler.submit(
        "proteinqc",
        params={
            "input_pdb": input_path,
            "tools": ",".join(tool_keys),
            "chains": ",".join(chains),
        },
        submission_args=dict(sync=True),
    )
    process.wait()
    if process.returncode == 0:
        console.print("[bold][green]✔[/green] ProteinQC completed successfully[/bold]")
        console.print("Example ProteinQC output saved to:")
        console.print(scheduler.get_output_dir(job_id))
    else:
        console.print("ProteinQC workflow [red]FAILED[/red], please see errors above")
        exit(process.returncode)


if __name__ == "__main__":
    app()



#
# Run everything including imports only in __main__ block
# This is done to avoid subprocesses initializing streamlit
#
import importlib
import os

if __name__ == "__main__":
    import streamlit as st
    from ovo import config, get_username
    from ovo.core.plugins import plugin_pages

    from ovo.app.pages import (
        welcome_page,
        jobs_page,
        designs_page,
        rf_scaffold_design_page,
        rf_binder_design_page,
        rf_binder_diversification_page,
        bindcraft_binder_design_page,
        admin_debug_page,
        admin_import_export_page,
    )

    pages = {
        "Browse": [
            welcome_page,
            jobs_page,
            designs_page,
        ],
        "RFdiffusion": [
            rf_scaffold_design_page,
            rf_binder_design_page,
            rf_binder_diversification_page,
        ],
        "BindCraft": [
            bindcraft_binder_design_page,
        ],
    }

    if plugin_pages:
        url_paths = set(page._url_path for group in pages.values() for page in group)
        for group_label, module_pages in plugin_pages.items():
            for page_dict in module_pages:
                assert isinstance(page_dict, dict), (
                    f"Expected dictionaries in '{module_name}.pages, found: {type(page_dict).__name__}: {page_dict}"
                )
                page_module_path = page_dict["page"]
                assert "/" not in page_module_path and not page_module_path.endswith(".py"), (
                    f"Plugin page should be defined by module path (my_tool.page_module), got {page_module_path}"
                )
                module_name = page_module_path.split(".")[0]
                module = importlib.import_module(module_name)
                url_path = page_module_path.split(".")[-1]
                while url_path in url_paths:
                    url_path += "_copy"
                url_paths.add(url_path)
                page_file_path = page_module_path.removeprefix(module_name).removeprefix(".").replace(".", "/") + ".py"
                pages[group_label] = pages.get(group_label, []) + [
                    st.Page(
                        page=os.path.join(os.path.dirname(module.__file__), page_file_path),
                        url_path=url_path,
                        **{k: v for k, v in page_dict.items() if k != "page"},
                    )
                ]

    if get_username() in config.auth.admin_users:
        pages["Admin section"] = [
            admin_debug_page,
            admin_import_export_page,
        ]

    pg = st.navigation(pages)
    pg.run()



import re
from dataclasses import dataclass, field
from typing import Callable, Optional

import pandas as pd

from ovo.core.database import descriptors_refolding, descriptors_rfdiffusion
from ovo.core.database.models import DesignWorkflow, WorkflowParams, WorkflowTypes, Design, Threshold, Base
from ovo.core.database.models_refolding import (
    RefoldingSupportedDesignWorkflow,
)
from ovo.core.scheduler.base_scheduler import Scheduler
from ovo.core.utils.residue_selection import (
    from_segments_to_hotspots,
    parse_partial_diffusion_binder_contig,
    create_partial_diffusion_binder_contig,
)

MODEL_WEIGHTS_SCAFFOLD = ["Base", "ActiveSite"]
MODEL_WEIGHTS_BINDER = ["Complex_base", "Complex_beta"]


@dataclass
class RFdiffusionParams(WorkflowParams):
    # path to input pdb in S3 or other storage backend
    input_pdb_paths: list[str] = field(default_factory=list, metadata=dict(show_to_user=False, storage_path=True))
    # contig strings
    contigs: list[str] = field(default_factory=list)
    # hotspot residues, comma-separated
    hotspots: str | None = None
    # model_weights
    model_weights: str | None = None
    # number of denoising timesteps
    timesteps: int = 50
    # use partial diffusion
    partial_diffusion: bool = False
    # number of backbone designs to generate
    num_designs: int = 10
    # enable cyclic offset to design macrocyclic peptides
    cyclic_offset: bool = False
    # Number of designs to be processed in batches
    batch_size: int = 100
    # Limit contig length to given value (123) or to given range (123-456, inclusive)
    contigmap_length: int | str | None = None
    # Inpaint (generate) sequence for these input structure regions, in this format: A10-20/B30-40/...
    inpaint_seq: str | None = None
    # Hard filters for backbone designs, applied before sequence design, e.g. "pydssp_helix_percent<50,N_contact_hotspots>=5"
    backbone_filters: str | None = None
    # Additional CLI params for RFdiffusion run_inference.py
    run_parameters: str = ""

    @property
    def input_pdb(self):
        return self.input_pdb_paths[0] if self.input_pdb_paths else None

    @input_pdb.setter
    def input_pdb(self, value):
        self.input_pdb_paths = [value]

    @property
    def contig(self):
        return self.contigs[0] if self.contigs else None

    @contig.setter
    def contig(self, value):
        self.contigs = [value] if value else []

    def validate(self):
        super().validate()
        if not self.input_pdb:
            raise ValueError("No input pdb provided")
        if not self.contig:
            raise ValueError("Please provide a contig")
        if "/0 " not in self.contig and " " in self.contig:
            raise ValueError(
                f'Spaces detected in contig specification, keep in mind that chain breaks are done by inserting "/0 ", found: "{self.contig}"'
            )
        if self.contigmap_length:
            assert isinstance(self.contigmap_length, (int, str)), (
                f"Expected int or str, got {self.contigmap_length} for contigmap_length"
            )
            assert (
                isinstance(self.contigmap_length, int)
                or re.fullmatch("[0-9]+", self.contigmap_length)
                or re.fullmatch("[0-9]+-[0-9]+", self.contigmap_length)
            ), f"Invalid contigmap_length, expected format 123 or 123-456, got: '{self.contigmap_length}'"
        if self.hotspots:
            assert isinstance(self.hotspots, str), f"Expected str for hotspots, got {type(self.hotspots).__name__}"
            if not all(re.fullmatch("[A-Z][0-9]+", hotspot) for hotspot in self.hotspots.split(",")):
                raise ValueError(f"Invalid hotspots format, expected 'A123,A124,A131', got: '{self.hotspots}'")
        if self.inpaint_seq:
            assert isinstance(self.inpaint_seq, str), (
                f"Expected str for inpaint_seq, got {type(self.inpaint_seq).__name__}"
            )
            if not all(re.fullmatch("[A-Z][0-9]+(-[0-9]+)?", segment) for segment in self.inpaint_seq.split("/")):
                raise ValueError(
                    f"Invalid inpaint_seq format, expected format 'A10-20/A22/B30-40', got: '{self.inpaint_seq}'"
                )
        if self.backbone_filters:
            assert isinstance(self.backbone_filters, str), (
                f"Expected str for backbone_filters, got {type(self.backbone_filters).__name__}"
            )
            for filter_expr in self.backbone_filters.split(","):
                match = re.fullmatch("([a-zA-Z0-9_]+)(<=|>=|=|<|>)([-]?[0-9]+(\\.[0-9]+)?)", filter_expr)
                if not match:
                    raise ValueError(
                        f"Invalid backbone_filters format, expected format 'pydssp_helix_percent<50,N_contact_hotspots>=5', got: '{self.backbone_filters}'"
                    )
                field, operator, value, _ = match.groups()
                if field not in descriptors_rfdiffusion.BACKBONE_METRIC_FIELD_NAMES:
                    raise ValueError(
                        f"Invalid backbone filter field '{field}', supported fields: {', '.join(descriptors_rfdiffusion.BACKBONE_METRIC_FIELD_NAMES)}"
                    )


@dataclass
class ProteinMPNNParams(WorkflowParams):
    # number of sequences to generate per structure
    num_sequences: int = 5
    # number of fastrelax cycles (0 = no fastrelax)
    fastrelax_cycles: int = 0
    # amino acids to omit, comma-separated
    omit_aa: str = "CX"
    # sampling temperature (0 = greedy, 0.1 = default, higher is more random)
    sampling_temp: float | None = field(default=None, metadata={"show_to_user": True})
    # amino acid bias, for example F:1.5,K:-1.5 (positive = more likely, zero = neutral default, negative = less likely)
    bias_aa: str = ""
    # additional commandline run parameters
    run_parameters: str = ""

    def validate(self):
        super().validate()
        if self.fastrelax_cycles:
            assert self.fastrelax_cycles >= 0, "fastrelax_cycles must be non-negative"
            if self.num_sequences != 1:
                # TODO implement support for multiple sequences per cycle with fastrelax
                raise NotImplementedError(
                    "Please set num_sequences=1 when using fastrelax_cycles > 0. "
                    "Generating multiple sequences per cycle with fastrelax is not supported yet."
                )
            # help user use proper run params to avoid failing after RFdiffusion stage
            if "--" in self.run_parameters:
                raise ValueError("Please use arguments for dl_interface_design_extended.py (single dash)")
        else:
            # help user use proper run params to avoid failing after RFdiffusion stage
            if self.run_parameters and self.run_parameters[0] == "-" and self.run_parameters[1] != "-":
                raise ValueError("Please use arguments for ligandmpnn (double dash)")
            assert self.num_sequences >= 1, "num_sequences must be at least 1"
        if not self.sampling_temp:
            raise ValueError("Please provide a sampling temperature")
        if self.bias_aa:
            assert isinstance(self.bias_aa, str), f"Expected str for bias_aa, got {type(self.bias_aa).__name__}"
            if not all(re.fullmatch("[A-Z]:-?[0-9]+(\\.[0-9]+)?", bias) for bias in self.bias_aa.split(",")):
                raise ValueError(f"Invalid bias_aa format, expected format 'F:1.5,K:-1.5', got: '{self.bias_aa}'")


@dataclass
class RefoldingParams(WorkflowParams):
    # Refolding test to perform, for example 'af2_model_1_ptm_tt_3rec'
    primary_test: str = None
    # run ESMFold with fp16 precision
    esmfold_fp16: bool = False

    def validate(self):
        super().validate()
        if not self.primary_test:
            raise ValueError("Please select a refolding test to perform")
        if "," in self.primary_test:
            raise ValueError(f"Only a single refolding test can be selected, got: {self.primary_test}")


@dataclass
class RFdiffusionWorkflow(DesignWorkflow, RefoldingSupportedDesignWorkflow):
    """Dataclass for RFdiffusion workflow."""

    # input parameters for RFdiffusion
    rfdiffusion_params: RFdiffusionParams | None = field(
        default_factory=RFdiffusionParams, metadata=dict(tool_name="RFdiffusion")
    )
    # input parameters for Protein MPNN
    protein_mpnn_params: ProteinMPNNParams | None = field(
        default_factory=ProteinMPNNParams, metadata=dict(tool_name="ProteinMPNN")
    )
    # input parameters for AlphaFold Initial Guess, ESMFold and other refolding test params
    refolding_params: RefoldingParams | None = field(
        default_factory=RefoldingParams, metadata=dict(tool_name="Refolding")
    )

    # UI state variables
    # User-provided PDB code (or filename in case of uploaded file)
    input_name: str | None = None
    # preview_job_id: job id of the submitted preview job
    preview_job_id: str | None = None
    # selected scaffolding segments or binder hotspots
    selected_segments: list[str] = None
    # IDs of designs used for partial diffusion (unless custom upload was used)
    # Corresponds to order of contigs in rfdiffusion_params.contigs field
    design_ids: list[str] = field(default_factory=list)

    def get_table_row(self, **kwargs) -> pd.Series:
        """Get all values of all param fields, skip fields with metadata.show_to_user=False, return pd.Series"""
        row = super().get_table_row(**kwargs)
        return pd.concat(
            [
                pd.Series(
                    {
                        ("Workflow", "input"): self.input_name,
                    },
                    **kwargs,
                ),
                row,
            ]
        )

    def get_input_pdb_path(self, contig_index=None):
        if contig_index is None:
            return self.rfdiffusion_params.input_pdb
        if contig_index == 0:
            return self.rfdiffusion_params.input_pdb_paths[0]
        has_multiple_inputs = len(self.rfdiffusion_params.input_pdb_paths) > 1
        has_multiple_contigs = len(self.rfdiffusion_params.contigs) > 1
        if has_multiple_inputs and has_multiple_contigs:
            assert len(self.rfdiffusion_params.input_pdb_paths) == len(self.rfdiffusion_params.contigs), (
                "When multiple input PDBs and multiple contigs are provided, "
                f"their counts must match. Got: {len(self.rfdiffusion_params.input_pdb_paths)} PDBs "
                f"and {len(self.rfdiffusion_params.contigs)} contigs."
            )
        if has_multiple_inputs:
            # When workflow has a single contig, the contig_index corresponds to PDB input index
            return self.rfdiffusion_params.input_pdb_paths[contig_index]
        else:
            # When workflow has multiple contigs, it must only have a single PDB input
            return self.rfdiffusion_params.input_pdb_paths[0]

    def get_contig(self, contig_index=0):
        return self.rfdiffusion_params.contigs[contig_index] if self.rfdiffusion_params.contigs else ""

    def set_contig(self, contig: str):
        self.rfdiffusion_params.contig = contig

    def get_hotspots(self):
        return self.rfdiffusion_params.hotspots

    def get_cyclic_offset(self):
        return self.rfdiffusion_params.cyclic_offset

    def get_selected_segments(self, inpainting=False):
        if inpainting:
            return self.rfdiffusion_params.inpaint_seq.split("/") if self.rfdiffusion_params.inpaint_seq else []
        else:
            return self.selected_segments or []

    def set_selected_segments(self, selection: list[str], inpainting=False):
        if inpainting:
            self.rfdiffusion_params.inpaint_seq = "/".join(selection) if selection else None
        else:
            self.selected_segments = selection

    def get_time_estimate(self, scheduler: Scheduler) -> str:
        total_designs = self.rfdiffusion_params.num_designs * len(self.rfdiffusion_params.contigs)
        STARTUP_TIME_MINUTES = scheduler.get_startup_time_minutes() * 3
        ONE_DESIGN_MINUTES = 2  # assuming that one full design takes this many minutes to complete
        estimated_minutes = STARTUP_TIME_MINUTES + total_designs * ONE_DESIGN_MINUTES
        estimated_hours = estimated_minutes / 60
        estimated_days = estimated_hours / 24

        if estimated_days > 1:
            final_estimation = f"{estimated_days:.1f} days"
        elif estimated_hours > 1:
            final_estimation = f"{estimated_hours:.1f} hours"
        else:
            final_estimation = f"{estimated_minutes:.1f} minutes"

        # FIXME improve time estimation based on chain length (quadratic?), number of contigs and batch size
        message = (
            f"The pipeline will generate and evaluate **{total_designs} backbone designs**. "
            f"For a structure of **~1000 residues**, the expected runtime is about **2 minutes** for 1 design. "
            f"The estimated total runtime including startup is **{final_estimation}**. "
        )
        if self.rfdiffusion_params.num_designs > self.rfdiffusion_params.batch_size:
            message += (
                f"Workflow will be parallelized in batches of {self.rfdiffusion_params.batch_size} designs, "
                f"each batch can run in parallel so the total runtime can be shortened dramatically."
            )
        return message

    def get_pipeline_name(self) -> str:
        return "ovo.rfdiffusion-end-to-end"

    def prepare_params(self, workdir: str) -> dict:
        from ovo.core.logic.design_logic_rfdiffusion import prepare_rfdiffusion_workflow_params

        return prepare_rfdiffusion_workflow_params(self, workdir=workdir)

    def process_results(self, job: "DesignJob", callback: Callable = None) -> list[Base]:
        """Process results of a successful workflow - download files from workdir,
        create Design and DescriptorValue objects"""
        from ovo.core.logic.design_logic_rfdiffusion import process_workflow_results

        return process_workflow_results(job=job, callback=callback)

    @classmethod
    def get_download_fields(cls):
        return {
            "RFdiffusion Input PDB": (RFdiffusionWorkflow, "rfdiffusion_params.input_pdb_paths"),
        }

    def get_refolding_design_paths(self, design_ids: list[str]) -> dict[str, str]:
        from ovo import db

        return {d.id: d.structure_path for d in db.select(Design, id__in=design_ids)}


@WorkflowTypes.register("RFdiffusion scaffold design")
@dataclass
class RFdiffusionScaffoldDesignWorkflow(RFdiffusionWorkflow):
    # acceptance threshold values (descriptor key -> interval (min, max, enabled))
    acceptance_thresholds: dict[str, Threshold] = field(
        default_factory=lambda: {
            descriptors_refolding.AF2_PRIMARY_PAE.key: Threshold(max_value=5.0),
            descriptors_refolding.AF2_PRIMARY_DESIGN_RMSD.key: Threshold(max_value=2.0),
            descriptors_refolding.AF2_PRIMARY_NATIVE_MOTIF_RMSD.key: Threshold(max_value=2.0),
            descriptors_refolding.AF2_PRIMARY_PLDDT.key: Threshold(min_value=80),
            descriptors_refolding.ESMFOLD_PAE.key: Threshold(max_value=10.0, enabled=False),
            descriptors_refolding.ESMFOLD_DESIGN_BACKBONE_RMSD.key: Threshold(max_value=5.0, enabled=False),
            descriptors_refolding.ESMFOLD_PLDDT.key: Threshold(min_value=80, enabled=False),
            descriptors_rfdiffusion.RADIUS_OF_GYRATION.key: Threshold(enabled=False),
        }
    )

    @classmethod
    def visualize_single_design_structures(cls, design_id: str):
        """Visualize single design structures in Streamlit"""
        from ovo.app.components.workflow_visualization_components import rfdiffusion_scaffold_design_visualization

        rfdiffusion_scaffold_design_visualization(design_id)

    @classmethod
    def visualize_single_design_sequences(self, design_id: str):
        from ovo.app.components.workflow_visualization_components import visualize_rfdiffusion_design_sequence

        visualize_rfdiffusion_design_sequence(design_id)

    def get_refolding_design_type(self) -> str:
        return "scaffold"

    def get_refolding_native_pdb_path(self, contig_index: int) -> str:
        # In binder design, we compare with fixed input motif from input PDB
        return self.get_input_pdb_path(contig_index=contig_index)


@WorkflowTypes.register("RFdiffusion binder design")
@dataclass
class RFdiffusionBinderDesignWorkflow(RFdiffusionWorkflow):
    # Workflow-specific parameters
    # length of the binder
    binder_length: str | None = None
    # dict chains, key: chain, value: contig, e.g. {A: A1-100/102-110/112-130, B: B23-202}
    chains: dict | None = None
    # start residue of the trimmed chain
    start_res_trimmed_chain: int | None = None
    # end residue of the trimmed chain
    end_res_trimmed_chain: int | None = None
    # chain ID of the target chain
    target_chain: str | None = None
    # acceptance threshold values (descriptor key -> interval (min, max, enabled))
    acceptance_thresholds: dict[str, Threshold] = field(
        default_factory=lambda: {
            descriptors_refolding.AF2_PRIMARY_IPAE.key: Threshold(max_value=10.0),
            descriptors_refolding.AF2_PRIMARY_TARGET_ALIGNED_BINDER_RMSD.key: Threshold(max_value=2.0),
            descriptors_refolding.AF2_PRIMARY_PLDDT_BINDER.key: Threshold(min_value=80),
            descriptors_rfdiffusion.PYROSETTA_DDG.key: Threshold(max_value=-30.0),
            descriptors_rfdiffusion.N_CONTACTS_TO_HOTSPOTS.key: Threshold(min_value=1, enabled=False),
            descriptors_refolding.AF2_PRIMARY_BINDER_PAE.key: Threshold(max_value=5.0, enabled=False),
            descriptors_rfdiffusion.PYROSETTA_CMS.key: Threshold(enabled=False),
            descriptors_rfdiffusion.PYROSETTA_SAP_SCORE.key: Threshold(enabled=False),
            descriptors_rfdiffusion.RADIUS_OF_GYRATION.key: Threshold(enabled=False),
        }
    )

    def get_target_chain(self):
        return self.target_chain

    def get_target_trim_boundary(self):
        return self.start_res_trimmed_chain, self.end_res_trimmed_chain

    def get_selected_segments(self, contig_index=0, partial_diffusion=False):
        if partial_diffusion:
            # get REDESIGNED segments from binder contig as they correspond to positions in binder chain A
            # default case for a peptide of 12 residues is simply binder_contig="12-12" turned into ["A1-12"]
            # or when redesigning only A1 and A3-7 in a peptide of 12 residues, this would be
            # binder_contig="1-1/A2-2/5-5/A8-12" turned into ["A1-1", "A3-7"]
            binder_contig = self.get_binder_contig(contig_index=contig_index)
            binder_length, designed_segments = parse_partial_diffusion_binder_contig(binder_contig)
            return designed_segments
        else:
            return super().get_selected_segments()

    def set_selected_segments(self, segments: list[str], contig_index=0, partial_diffusion=False):
        if partial_diffusion:
            # create contig from REDESIGNED segments as they correspond to positions in binder chain A
            # default case for a peptide of 12 residues:
            # segments=[] -> binder_contig="12-12"
            # or when redesigning only A1 and A3-7 in a peptide of 12 residues,
            # segments=["A1-1", "A3-7"] -> binder_contig="1-1/A2-2/5-5/A8-12"
            binder_length, _ = parse_partial_diffusion_binder_contig(self.get_binder_contig(contig_index=contig_index))
            self.set_binder_contig(
                create_partial_diffusion_binder_contig(redesigned_segments=segments, binder_length=binder_length)
            )
        else:
            # Hotspots
            super().set_selected_segments(segments)
            # Convert from segments to hotspots
            self.rfdiffusion_params.hotspots = from_segments_to_hotspots(segments)

    def get_target_contig(self, contig_index=0):
        contig = self.get_contig(contig_index=0)
        if not contig:
            return ""
        subcontigs = contig.split()
        # make sure that target contig is fixed
        assert len(subcontigs) == 2, f"Expected a binder chain and target chain in contig, got: {contig}"
        assert all(segment[0].isalpha() for segment in subcontigs[1].split("/")), (
            f"Expected contig in format 'DESIGN/0 TARGET', got: {contig}"
        )
        return subcontigs[1]

    def get_binder_contig(self, contig_index=0):
        contig = self.get_contig(contig_index=0)
        if not contig:
            return ""
        subcontigs = contig.split()
        assert len(subcontigs) == 2, f"Expected a binder chain and target chain in contig, got: {contig}"
        # make sure that target contig is fixed (binder contig should be designed but not necessarily - e.g. in partial diffusion)
        assert all(segment[0].isalpha() for segment in subcontigs[1].split("/")), (
            f"Expected contig in format 'DESIGN/0 TARGET', got: {contig}"
        )
        return subcontigs[0].removesuffix("/0")

    def set_binder_contig(self, binder_contig: str, contig_index: int = 0):
        target_contig = self.get_target_contig()
        assert target_contig, "Target contig must be set before setting binder contig"
        self.rfdiffusion_params.contigs[contig_index] = binder_contig + "/0 " + target_contig

    @classmethod
    def visualize_multiple_designs_structures(cls, design_ids: list[str]):
        """Visualize multiple design structures in Streamlit"""
        from ovo.app.components.workflow_visualization_components import (
            rfdiffusion_multiple_binder_designs_visualization,
        )

        rfdiffusion_multiple_binder_designs_visualization(design_ids)

    @classmethod
    def visualize_single_design_structures(cls, design_id: str):
        """Visualize single design structures in Streamlit"""
        from ovo.app.components.workflow_visualization_components import rfdiffusion_binder_design_visualization

        rfdiffusion_binder_design_visualization(design_id)

    def get_refolding_design_type(self) -> str:
        return "binder"

    def get_refolding_native_pdb_path(self, contig_index: int) -> Optional[str]:
        # In binder design, we don't compare with fixed input motif, so return None
        return None



import argparse
import os
import glob
import json

import numpy as np
import pandas as pd
import string

from biopandas.pdb import PandasPdb
from copy import deepcopy

# CONTIG: A1-10/10-20/A30-40/0 B23-40/.../0 C10-25
# CONTIG_CHAINS: list or comma-separated list = A1-10/10-20/A30-40/0, B23-40/.../0, C10-25
# MASKED_CONTIG_CHAINS: CONTIG_CHAINS read from trb file
# SEGMENTS: list or comma-separated list = A1-10, 10-20, A30-40, B23-40, ...

ALPHABET = string.ascii_uppercase


def is_designed(contig_chain_or_segment: str) -> bool:
    """
    Check if the contig chain is fixed or not.
    :param contig_chain: str, the contig chain string
    :return: bool, True if designed, False if is fixed
    """
    for segment in contig_chain_or_segment.removesuffix("/0").split("/"):
        if not segment[0].isalpha():
            return True
    return False


def order_designed_contig(contig_chains: list[str], trb_ref_pdb_idx: list[tuple[str, str]]) -> list[str]:
    """
    Order the contig segments by designed and fixed segments. It is needed to guarantee the correct mapping with
    the output pdb numbering as in the output pdb designed input residues are first listed in A,
    then fixed residues listed in B.
    First, create a list with designed + fixed contig segments.
    Then, scans the segments and check if the segments are ordered correctly comparing with the rfdiffusion output.
    Example:
        - input contig_chains [A1-10/10-20/A30-40/0, B23-40/.../0, C10-25] (designed + fixed)
        - For each contig segment splits it in segments: i.e. A1-10, 10-20, A30-40, B23-40, ...
            For fixed segments (namely A1-10, A30-40, B23-40) check if the segments are ordered as in the trb ref pdb
            idx.
            This latter gets only fixed segments trb_ref_pdb_idx:
            [(A, 1), (A, 2), ... (A, 10), (A, 30), (A, 31), ..., (B, 23), (B, 24), ..., (B, 40), ..., (C, 10), (C, 11)
            ...]
    """
    designed_segments, fixed_segments = [], []

    for contig_chain in contig_chains:
        if is_designed(contig_chain):
            designed_segments.append(contig_chain)
        else:
            fixed_segments.append(contig_chain)

    ordered_contig_chains = designed_segments + fixed_segments

    # SANITY CHECK
    # Check if the segments are ordered correctly comparing with the rfdiffusion output
    trb_mask_idx = 0
    for ordered_contig_chain in ordered_contig_chains:
        for segment in ordered_contig_chain.removesuffix("/0").split("/"):
            # Check if segment is designed
            if is_designed(segment):
                continue

            chain = segment[0]
            start_resnum, end_resnum = map(int, segment[1:].split("-"))
            for res_num in range(start_resnum, end_resnum + 1):
                if (chain, res_num) != trb_ref_pdb_idx[trb_mask_idx]:
                    raise ValueError(
                        f"Mismatch between parsed residue {(chain, res_num)} and "
                        f"expected rf-diffusion output {trb_ref_pdb_idx[trb_mask_idx]} in the {trb_mask_idx} "
                        f"trb mask location: check 'trb['complex_con_ref_pdb_idx']'"
                    )
                trb_mask_idx += 1

    return ordered_contig_chains


def input_output_standard_mapping(
    contig_chains: list[str],
) -> tuple[list[tuple[str, str] | None], list[tuple[str, str]], list[tuple[str, str]]]:
    """Create the input, output and standardized numbering for the PDB file
    :param contig_segments: list of str with the contig segments ordered by order_designed_contig
        [A1-10/10-20/A30-40/0, B23-40/.../0, C10-25]
        [chain designed, chain designed, chain fixed]
    :return: tuple of three lists with the input, output and standardized numbering
        Split each chain into segments.
        If the segment is designed, the input numbering is None (there are no correspondence in the input)
        Rf diffusion output pdb numbering is consecutive starting from 1 (guaranteed beacuse of the order of the segments)
        designed chains are in 'A' and fixed chain(s) are in 'B'.
        Standardized numbering include:
            - Consecutive chain id, new_chain := ALPHABET[i] where 'i' runs with the contig (chains) segments,
                each chain break is a different chain.
            - Standard numbering: use the input pdb numbering if chain is fixed,
                otherwise use consecutive numbering starting from 1 in each chain
        Example:
            Input pdb numbering:
            First designed chain                                                Second designed chain  Third (fixed) chain
            [(A, 1), (A, 2), ..., None   , None   , ..., (A, 30), (A, 31), ..., (B, 23), (B, 24), ..., (C, 10), (C, 11), ...]
            Rf diffusion output pdb numbering:
            [(A, 1), (A, 2), ..., (A, 11), (A, 12), ..., (A, 30), (A, 31), ..., (A, n), (B, n+1), ..., (B, n+k), (B, n+k+1), ...]
            Standardized pdb numbering:
            [(A, 1), (A, 2), ..., (A, 11), (A, 12), ..., (A, 30), (A, 31), ..., (B, 1), (B, 2), ..., (C, 10), (C, 11), ...]
    """

    input_numbering = []
    rfdiff_output_numbering = []
    standardized_numbering = []
    rfdiff_output_num = 1
    for i, contig_chain in enumerate(contig_chains):
        rfdiff_output_chain = "A" if is_designed(contig_chain) else "B"

        new_chain = ALPHABET[i]
        standard_resnum = 1

        for segment in contig_chain.removesuffix("/0").split("/"):
            if is_designed(segment):
                start_len, end_len = map(int, segment.split("-"))
                assert start_len == end_len, (
                    f"Start len and end len of design segment should coincide: "
                    f"{start_len} != {end_len} in {contig_chain} "
                )

                for _ in range(start_len):
                    input_numbering.append(None)
                    rfdiff_output_numbering.append((rfdiff_output_chain, str(rfdiff_output_num)))
                    standardized_numbering.append((new_chain, str(standard_resnum)))
                    standard_resnum += 1
                    rfdiff_output_num += 1
            else:
                input_chain = segment[0]
                start_resnum, end_resnum = map(int, segment[1:].split("-"))

                for input_num in range(start_resnum, end_resnum + 1):
                    input_numbering.append((input_chain, str(input_num)))
                    rfdiff_output_numbering.append((rfdiff_output_chain, str(rfdiff_output_num)))
                    if rfdiff_output_chain == "A":
                        standardized_numbering.append((new_chain, str(standard_resnum)))
                        standard_resnum += 1
                    else:
                        standardized_numbering.append((new_chain, str(input_num)))
                    rfdiff_output_num += 1

    return input_numbering, rfdiff_output_numbering, standardized_numbering


def chunk_string(s: str | None, chunk_size: int = 40) -> str | None:
    """
    Splits the input string `s` into chunks of size `chunk_size`.
    Args:
        s (str or None): The input string to be chunked.
        chunk_size (int): The maximum length of each chunk.
    Yields:
        str: A chunk of the input string, or returns None if input is None.
    """
    if s is None:
        return None

    for i in range(0, len(s), chunk_size):
        yield s[i : i + chunk_size]


def renumber_rfdiffusion_pdb(pdb_file: str, trb_file: str, output_pdb: str):
    """Renumber the RF-diffusion target chain residues in the input PDB file
    :param:
        pdb_file: str with the path to the rfdiffusion PDB file
        trb_file: str with the path to the rfdiffusion TRB file
        output_pdb: str with the path to save the standardized PDB file

    """

    trb = pd.read_pickle(trb_file)
    ppdb = PandasPdb().read_pdb(pdb_file)

    # Adding hotspot in a REMARK line
    renum_ppdb = deepcopy(ppdb)

    # # trb sampled_mask contains the exact length of generated segments in case of ranged input
    masked_contig_chains = trb["sampled_mask"]

    if "complex_con_ref_pdb_idx" in trb:
        trb_ref_pdb_idx = trb["complex_con_ref_pdb_idx"]
        # Order first designed chains and after fixed chains
        ordered_contig_chains = order_designed_contig(masked_contig_chains, trb_ref_pdb_idx)
        standardized_contig = " ".join(ordered_contig_chains)
    else:
        # If no complex_con_ref_pdb_idx, use the masked segments as is
        ordered_contig_chains = masked_contig_chains
        standardized_contig = ordered_contig_chains[0]

    input_num, output_num, standard_num = input_output_standard_mapping(ordered_contig_chains)

    for (out_chain, out_resnum), (standard_chain, standard_res_num) in zip(output_num, standard_num):
        query_resnum = int(out_resnum)

        queried_ids = ppdb.df["ATOM"].query("residue_number==@query_resnum & chain_id==@out_chain").index

        assert ppdb.df["ATOM"].residue_number.dtype == type(query_resnum) == int, (
            f"Unexpected error standardizing PDB, residue_number should be int: {pdb_file} "
            f"residue_number=={ppdb.df['ATOM'].residue_number.dtype} != {type(query_resnum)}"
        )

        assert len(queried_ids), (
            f"Unexpected error standardizing PDB, residue not found in {pdb_file}: "
            f"residue_number=={query_resnum} chain_id=={out_chain}"
        )

        renum_ppdb.df["ATOM"].loc[queried_ids, "residue_number"] = [standard_res_num] * len(queried_ids)
        renum_ppdb.df["ATOM"].loc[queried_ids, "chain_id"] = [standard_chain] * len(queried_ids)

    # Get input contig and hotspots format for standardized pdb
    input_contig = trb["config"]["contigmap"]["contigs"][0]
    input_hotspots_list = trb["config"]["ppi"]["hotspot_res"]
    input_hotspots = ",".join(input_hotspots_list) if input_hotspots_list else None

    # Standardize the hotspots
    out_hotspots = []
    if input_hotspots_list:
        for hotspot in input_hotspots_list:
            chain, hotspot_num = hotspot[0], hotspot[1:]
            if (chain, hotspot_num) not in input_num:
                raise ValueError(f"Hotspot {hotspot} not valid for structure {pdb_file}")
            res_tuple = standard_num[input_num.index((chain, hotspot_num))]
            out_hotspots.append("".join(res_tuple))

    standardized_hotspots = ",".join(out_hotspots) if out_hotspots else None

    # Get designed and fixed chains
    chains = " ".join([ALPHABET[i] for i, _ in enumerate(ordered_contig_chains)])

    # Get inpaint sequence if available
    inpaint_seq = (
        "/".join(trb["config"]["contigmap"]["inpaint_seq"]) if trb["config"]["contigmap"]["inpaint_seq"] else None
    )

    # Write REMARK lines in the standardized pdb
    # TODO: Consider using some code 1070 (only int allowed)
    for chunk in chunk_string(input_contig, 40):
        renum_ppdb.add_remark(code=1, text=f" Input contig: {json.dumps(chunk)}")

    for chunk in chunk_string(standardized_contig, 40):
        renum_ppdb.add_remark(code=1, text=f" Standardized contig: {json.dumps(chunk)}")

    if inpaint_seq:
        for chunk in chunk_string(inpaint_seq, 40):
            renum_ppdb.add_remark(code=1, text=f" Inpaint seq: {json.dumps(chunk)}")

    for chunk in chunk_string(chains, 40):
        renum_ppdb.add_remark(code=1, text=f" Chains: {json.dumps(chunk)}")

    if not input_hotspots:
        renum_ppdb.add_remark(code=1, text=" Input hotspots: ")

    for chunk in chunk_string(input_hotspots, 40):
        renum_ppdb.add_remark(code=1, text=f" Input hotspots: {json.dumps(chunk)}")

    if not standardized_hotspots:
        renum_ppdb.add_remark(code=1, text=" Standardized hotspots: ")

    for chunk in chunk_string(standardized_hotspots, 40):
        renum_ppdb.add_remark(code=1, text=f" Standardized hotspots: {json.dumps(chunk)}")

    renum_ppdb.to_pdb(output_path)

    # Add FIXED remarks for fastrelax protocol
    if "receptor_con_hal_pdb_idx" in trb:
        # Identify the last residue number in A chain
        last_res_id = int(trb["receptor_con_hal_pdb_idx"][0][1]) - 1

        # Identify where inpaint seq is True, ie, kept fixed
        indices = np.where(trb["inpaint_seq"][:last_res_id])[0]
    else:
        # Identify where inpaint seq is True in the only chain
        indices = np.where(trb["inpaint_seq"])[0]

    with open(output_pdb, "a") as f:
        f.write("\n")
        for position in indices:
            f.write(f"REMARK PDBinfo-LABEL:{position + 1: >5} FIXED\n")


if __name__ == "__main__":
    parser = argparse.ArgumentParser()
    parser.add_argument("input_pdb_path", type=str)
    parser.add_argument("input_trb_path", type=str)
    parser.add_argument("output_pdb_path", type=str)
    options = parser.parse_args()

    if os.path.isdir(options.input_pdb_path):
        pdb_paths = sorted(glob.glob(os.path.join(options.input_pdb_path, "*.pdb")))
        trb_paths = sorted(glob.glob(os.path.join(options.input_trb_path, "*.trb")))
    else:
        pdb_paths = [options.input_pdb_path]
        trb_paths = [options.input_trb_path]

    # Create sets to hold the base names (without extensions)
    pdb_basenames = {os.path.splitext(os.path.basename(pdb))[0] for pdb in pdb_paths}
    trb_basenames = {os.path.splitext(os.path.basename(trb))[0] for trb in trb_paths}

    # Find paired and unpaired files
    paired_filename, unpaired_pdb, unpaired_trb = [], [], []
    for basename in pdb_basenames:
        if basename in trb_basenames:
            paired_filename.append(
                (os.path.join(options.input_pdb_path, basename), os.path.join(options.input_trb_path, basename))
            )
        else:
            unpaired_pdb.append(basename)

    for basename in trb_basenames:
        if basename not in pdb_basenames:
            unpaired_trb.append(basename)

    assert not unpaired_pdb and not unpaired_trb, (
        f"Unpaired pdb files {unpaired_pdb}, trb files {unpaired_trb} found. "
        f"Please check the input directories {options.input_pdb_path}, {options.input_trb_path}."
    )

    assert paired_filename, "No paired pdb and trb files found in the input directories."

    print(f"Renumbering and standardizing {len(paired_filename)} PDB files.")
    for pdb_name, trb_name in paired_filename:
        output_path = os.path.join(options.output_pdb_path, os.path.basename(pdb_name) + "_standardized.pdb")
        renumber_rfdiffusion_pdb(pdb_name + ".pdb", trb_name + ".trb", output_path)
    print("Saved to:", options.output_pdb_path)



import json
import signal
import subprocess
import shlex
from pathlib import Path
from typing import Any
from uuid import uuid1
import importlib
import pandas as pd
import psutil
import os
import re
from datetime import datetime
from ovo.core.scheduler.base_scheduler import JobNotFound, Scheduler, SchedulerTypes
from ovo.core.scheduler.simple_queue_mixin import SimpleQueueMixin
from ovo.cli.common import init_nextflow, run_nextflow, OVOCliError
from ovo.core.utils.param_validation import validate_params, flatten_schema
import shutil


@SchedulerTypes.register()
class NextflowScheduler(Scheduler, SimpleQueueMixin):
    def submit(self, pipeline_name: str, params: dict = None, submission_args: dict = None) -> str:
        """
        Submits a Nextflow workflow asynchronously and returns the PID.

        :param pipeline_name: Workflow name and revision (ovo.rfdiffusion-end-to-end or github url with @version)
        :param params: Dictionary of parameters to pass to the workflow.
        :param submission_args: Submission arguments for the scheduler, overrides values in self.submission_args
        :return: Scheduler job ID
        """
        if not self.allow_submit:
            raise RuntimeError("Job submission is disabled")
        submission_args = {**self.submission_args, **(submission_args or {})}
        sync = submission_args.pop("sync", False)
        if self.has_queue():
            # When using queue workers, do not use -bg mode
            submission_args.pop("queue", None)
            sync = True

        command = self._create_run_command(
            pipeline_name=pipeline_name,
            params=params,
            sync=sync,
            submission_args=submission_args,
        )

        # Make sure nextflow is initialized to avoid unexpected failures later
        init_nextflow()

        job_id = str(uuid1())  # uuid1 is increasing by time
        execdir = self._get_exec_dir(job_id)
        os.makedirs(self.workdir, exist_ok=True)
        os.makedirs(execdir, exist_ok=False)  # fail in rare case that uuid1 is not unique

        print(f"Submitting workflow: {shlex.join(command)}")
        print(f"Execution directory: {execdir}")

        if self.has_queue():
            # Submit the command to the queue. Worker will run queue_run_task() in the worker loop.
            self.queue_put(job_id, command)
            return job_id
        else:
            # Run the command directly in a subprocess
            return self._run_subprocess(command, job_id, sync=sync)

    def _run_subprocess(self, command: list[str], job_id: str, sync: bool = False):
        # Run the Nextflow command in a child process
        out = None if sync else subprocess.DEVNULL
        env = {
            # inherit all environment variables
            **os.environ,
            # NOTE this forces conda to *prepend* the activated environment to PATH instead of replacing.
            # This avoids an issue which happens when a virtualenv with OVO is activated on top of a conda env,
            # where the virtualenv's bin directory is prepended to PATH, and conda env activation
            # only replaces the original env with the new env but doesn't change the order, so virtualenv takes precedence.
            "CONDA_SHLVL": "0",
        }
        execdir = self._get_exec_dir(job_id)
        process = subprocess.Popen(command, cwd=execdir, stdout=out, stderr=out, env=env)

        # Write down the process ID into .nextflow.pid file
        # Note that when running with sync=False, this file will be replaced by Nextflow with new background PID
        with open(os.path.join(execdir, ".nextflow.pid"), "w") as f:
            f.write(str(process.pid))

        if sync:
            process.wait()
            return process, job_id

        return job_id

    def queue_run_task(self, job_id: str, task: Any):
        """Execute a single task from the queue synchronously - executed in the worker loop"""
        # Run synchronously, print to stdout and stderr, ignore exit codes
        assert isinstance(task, list), (
            f"Expected task to be a list of command arguments, got: {task} ({type(task).__name__})"
        )
        return self._run_subprocess(task, job_id, sync=True)

    def run(
        self,
        pipeline_name: str,
        output_dir: str = None,
        params: dict = None,
        submission_args: dict = None,
        link=False,
    ) -> str:
        """Run workflow, wait for it to finish and return (process, job_id) tuple.

        Raise RuntimeError if the workflow fails (return code != 0).

        Shorthand for submit(submission_args=dict(sync=True))

        :param pipeline_name: Workflow name and revision (ovo.rfdiffusion-end-to-end or github url with @version)
        :param output_dir: output directory to save results to (optional, will be symlinked from workdir/execdir/job_id/output)
        :param params: Dictionary of parameters to pass to the workflow.
        :param submission_args: Submission arguments for the scheduler, overrides values in self.submission_args
        :param link: If True, symlink the output directory to output_dir instead of copying (only if output_dir is set)
        :return job_id: Scheduler job ID
        """
        if output_dir is not None:
            assert isinstance(output_dir, str) or isinstance(output_dir, Path), (
                f"output_dir should be a string or Path, got {type(output_dir).__name__}"
            )
            if os.path.exists(output_dir):
                raise FileExistsError(
                    f"Output directory already exists, please remove it or choose another directory: {output_dir}"
                )
        process, job_id = self.submit(
            pipeline_name, params=params, submission_args={**(submission_args or {}), "sync": True}
        )
        if process.returncode != 0:
            raise RuntimeError(
                f"{pipeline_name} failed with return code {process.returncode}. "
                f"See {self._get_exec_dir(job_id)}/.nextflow.log"
            )
        if output_dir is not None:
            os.makedirs(os.path.dirname(output_dir.rstrip("/")), exist_ok=True)
            if link:
                os.symlink(self.get_output_dir(job_id), output_dir)
            else:
                shutil.copytree(self.get_output_dir(job_id), output_dir)
        return job_id

    def _create_run_command(
        self, pipeline_name: str, params: dict = None, sync: bool = False, submission_args: dict = None
    ) -> list[str]:
        """
        Prepare a Nextflow run command for job submission.

        :param pipeline_name: Workflow name and revision (ovo.rfdiffusion-end-to-end or github url with @version)
        :param params: Dictionary of parameters to pass to the workflow.
        :param sync: If true, run the workflow synchronously.
        :param submission_args: Submission arguments for the scheduler, overrides values in self.submission_args
        :return: Command as a list of strings
        """
        assert isinstance(params, dict), f"params should be a dictionary, got: {type(params).__name__}"

        submission_args = submission_args.copy()
        max_memory = submission_args.pop("max_memory", None)
        resolve_relative_paths = submission_args.pop("resolve_relative_paths", True)

        pipeline_dir = self._get_pipeline_dir(pipeline_name)

        command = ["nextflow", "run"]
        command += ["-with-trace", "trace.txt"]
        command += ["-with-report", "report.html"]
        command += ["-work-dir", os.path.join(self.workdir, "work")]
        command += [pipeline_dir]
        command += ["--publish_dir", "output"]
        command += ["--reference_files_dir", self.reference_files_dir]
        command += ["--shared_modules", self._get_shared_modules_string()]
        command += ["-config", self._get_default_config_path()]

        if os.path.exists(os.path.join(pipeline_dir, "nextflow.config")):
            # Explicitly add pipeline's nextflow.config to override our default config
            command += ["-config", os.path.join(pipeline_dir, "nextflow.config")]

        for arg, value in submission_args.items():
            if isinstance(value, bool):
                value = str(value).lower()
            command += [f"-{arg}", str(value)]

        if max_memory:
            command += ["--max_memory", max_memory]

        if not sync:
            command.extend(["-ansi-log", "false"])
            command.extend(["-bg"])

        # Add workflow parameters
        if params:
            # Validate parameters against workflow schema if available
            schema = self.get_param_schema(pipeline_name)
            if schema:
                validate_params(params, schema)

            for key, value in params.items():
                if isinstance(value, str) and not value.strip():
                    # empty strings are interpreted as "true" by nextflow, better to skip that arg altogether
                    continue
                if value is None:
                    continue
                if isinstance(value, bool):
                    value = str(value).lower()
                if isinstance(value, str) and os.path.exists(value) and not os.path.isabs(value):
                    if resolve_relative_paths:
                        # Convert relative path to absolute path
                        abspath = os.path.abspath(value)
                        if value.endswith("/") and not abspath.endswith("/"):
                            # Make sure trailing slash is preserved
                            abspath += "/"
                        print(f"\nResolved relative path for parameter --{key}: {value} -> {abspath}\n")
                        value = abspath
                    else:
                        # We don't want to fail here but we can at least show a warning
                        print(f"Passing relative paths is not supported:\n{value}")
                command.append(f"--{key}")
                if isinstance(value, Path):
                    command.append(str(value.resolve()))
                else:
                    command.append(str(value))

        return command

    def _get_module_path(self, module_name) -> str:
        """Get absolute path to modulename directory"""
        module = importlib.import_module(module_name)
        module_path = os.path.abspath(os.path.dirname(module.__file__))
        return module_path

    def _get_default_config_path(self):
        """Get path to nextflow_default.config file"""
        module_path = self._get_module_path("ovo")
        return os.path.join(module_path, "pipelines/nextflow_default.config")

    def _get_pipeline_dir(self, pipeline_name: str) -> str:
        """Get path to workflow directory (with main.nf, nextflow.config, etc)"""
        if re.fullmatch(r"https?://.*", pipeline_name):
            if "@" not in pipeline_name:
                raise ValueError(f"Workflow URL must contain @revision suffix, got: {pipeline_name}")
            url, revision = pipeline_name.split("@", 1)
            info_data = None
            current_revision = None
            try:
                # Try getting workflow info to speed up the case when it already exists
                info_data = run_nextflow(["info", url, "-o", "json"])
                info = json.loads(info_data)
                current_revision = info["revisions"]["current"]
            except OVOCliError:
                pass
            if not info_data or not current_revision or revision != current_revision:
                # Pull workflow only if not already available or not checked out to the right revision
                run_nextflow(["pull", url, "-r", revision])
                info_data = run_nextflow(["info", url, "-o", "json"])
            info = json.loads(info_data)
            pipeline_dir = info["localPath"]
        elif "/" in pipeline_name or pipeline_name == ".":
            if not os.path.exists(pipeline_name):
                raise FileNotFoundError(f"Workflow path not found: {pipeline_name}")
            pipeline_dir = os.path.abspath(pipeline_name)
        elif pipeline_name.endswith(".nf"):
            if not os.path.exists(pipeline_name):
                raise FileNotFoundError(f"Workflow path not found: {pipeline_name}")
            pipeline_dir = os.path.abspath(os.path.dirname(pipeline_name))
        else:
            if "." in pipeline_name:  # custom workflow path (python_module_name.my_workflow)
                module_name, pipeline_subpath = pipeline_name.split(".")
            else:
                module_name = "ovo"
                pipeline_subpath = pipeline_name
            module_path = self._get_module_path(module_name)
            pipeline_dir = os.path.join(module_path, "pipelines", pipeline_subpath)
            if not os.path.isdir(pipeline_dir):
                raise FileNotFoundError(
                    f"Pipeline {pipeline_name} not available (directory not found: {pipeline_dir}). "
                    "Available pipelines: " + ", ".join(self.get_pipeline_names())
                )
            main_workflow_file_path = os.path.join(pipeline_dir, "main.nf")
            if not os.path.exists(main_workflow_file_path):
                raise FileNotFoundError(
                    f"Workflow {pipeline_name} not available (workflow main.nf file not found: {main_workflow_file_path}). "
                )
        return pipeline_dir

    def get_trace_table(self, job_id: str) -> pd.DataFrame | None:
        execdir = self._get_exec_dir(job_id)
        trace_file = os.path.join(execdir, "trace.txt")
        if not os.path.exists(trace_file):
            return None
        return pd.read_csv(trace_file, sep="\t", index_col=0)

    def get_status_label(self, job_id: str) -> str:
        """Get human-readable job status label. Should NOT be used to determine job status."""
        result = self.get_result(job_id)
        if result is None:
            try:
                self._get_pid(job_id)
            except JobNotFound:
                return "Queued"
            return "Running"
        return "Done" if result else "Failed"

    def get_result(self, job_id: str) -> bool | None:
        """Get job result: True if successful, False if failed, None if still running."""
        execdir = self._get_exec_dir(job_id)
        if not os.path.exists(execdir):
            raise JobNotFound(f"Job {job_id} execution directory not found: {execdir}")
        history_file = os.path.join(execdir, ".nextflow", "history")
        if os.path.exists(history_file):
            with open(history_file, "r") as f:
                lines = f.readlines()
                if not lines:
                    return None
                if lines:
                    # Use last history entry in case of multiple retries
                    fields = lines[-1].strip().split("\t")
                    status = fields[3]
                    assert status in ["-", "OK", "ERR"], f"Unexpected job status '{status}' in: {history_file}"
                    if status == "OK":
                        return True
                    elif status == "ERR":
                        return False
        try:
            pid = self._get_pid(job_id)
        except JobNotFound:
            if self.has_queue():
                # When using queue workers, we cannot reliably determine if the job is still running,
                # we need to rely on the worker to update the history file or the pid file
                return None
            raise
        if not psutil.pid_exists(pid):
            print(f"Job {job_id} process PID {pid} doesn't exist anymore, assuming job failed")
            return False
        # if process with given PID still exists, assume still running
        return None

    def get_log(self, job_id: str) -> str | None:
        """Get job execution log"""
        execdir = self._get_exec_dir(job_id)
        log_file = os.path.join(execdir, ".nextflow.log")
        if not os.path.exists(log_file):
            return None
        with open(log_file, "r") as f:
            return f.read()

    def cancel(self, job_id):
        """Cancel job execution"""
        pid = self._get_pid(job_id)
        signum = signal.SIGTERM
        os.kill(pid, signum)
        print("Killing", pid, "with signal", signum)

    def get_output_dir(self, job_id: str):
        """Get job output path"""
        execdir = self._get_exec_dir(job_id)
        return os.path.join(execdir, "output")

    def _get_exec_dir(self, job_id: str):
        if not isinstance(job_id, str):
            raise ValueError(
                f"Job ID of {type(self).__name__} should be a string, got {job_id} ({type(job_id).__name__}). Make sure to use job_id instead of the DB id."
            )
        execdir = os.path.join(self.workdir, "execdir", job_id)
        return execdir

    def _get_pid(self, job_id) -> int:
        execdir = self._get_exec_dir(job_id)
        if not os.path.exists(execdir):
            raise JobNotFound(f"Job {job_id} execution directory not found: {execdir}")
        pid_file = os.path.join(execdir, ".nextflow.pid")
        if not os.path.exists(pid_file):
            raise JobNotFound(f"Job {job_id} PID file not found: {pid_file}")
        with open(pid_file, "r") as f:
            pid = int(f.read())
        return pid

    def get_job_start_time(self, job_id: str) -> datetime | None:
        """Get job start time"""
        log = self.get_log(job_id)
        if not log:
            return None

        for line in log.splitlines():
            # Regex to match timestamps
            match_start = re.search(
                r"(\w{3}-\d{1,2} \d{2}:\d{2}:\d{2}\.\d{3}) \[main\] DEBUG nextflow.Session - Session start", line
            )
            if match_start:
                start_time = match_start.group(1)
                local_date = datetime.strptime(f"{datetime.now().year} {start_time}", "%Y %b-%d %H:%M:%S.%f")
                return datetime.utcnow() - datetime.now() + local_date
        return None

    def get_job_stop_time(self, job_id: str) -> datetime | None:
        """Get job end time"""
        log = self.get_log(job_id)
        if not log:
            return None

        for line in log.splitlines():
            # Regex to match timestamps
            match_stop = re.search(
                r"(\w{3}-\d{1,2} \d{2}:\d{2}:\d{2}\.\d{3}) \[main\] DEBUG nextflow.Session - Session completed", line
            )
            if match_stop:
                stop_time = match_stop.group(1)
                local_date = datetime.strptime(f"{datetime.now().year} {stop_time}", "%Y %b-%d %H:%M:%S.%f")
                return datetime.utcnow() - datetime.now() + local_date
        return None

    def _get_shared_module_paths(self):
        from ovo.core.plugins import plugin_modules

        module_names = ["ovo"] + sorted(set([m.split(".")[0] for m in plugin_modules]))
        return [(module_name, self._get_module_path(module_name)) for module_name in module_names]

    def _get_shared_modules_string(self) -> str:
        return ",".join(f"{module_name}:{module_path}" for module_name, module_path in self._get_shared_module_paths())

    def get_pipeline_names(self) -> list[str]:
        pipeline_names = []
        for module_name, module_path in self._get_shared_module_paths():
            pipelines_dir = os.path.join(module_path, "pipelines")
            if not os.path.isdir(pipelines_dir):
                continue
            for dirname in sorted(os.listdir(pipelines_dir)):
                if os.path.exists(os.path.join(pipelines_dir, dirname, "main.nf")):
                    pipeline_names.append(dirname if module_name == "ovo" else f"{module_name}.{dirname}")
        return pipeline_names

    def get_param_schema(self, pipeline_name: str) -> dict | None:
        """Get workflow schema JSON dict (JSON Schema standard) or None if not available."""
        pipeline_dir = self._get_pipeline_dir(pipeline_name)
        schema_path = os.path.join(pipeline_dir, "nextflow_schema.json")
        if not os.path.exists(schema_path):
            return None
        with open(schema_path, "r") as f:
            schema = json.load(f)
        # NOTE: always consider additionalProperties to be false
        schema["additionalProperties"] = False
        if "properties" not in schema:
            assert "allOf" in schema, f"Invalid schema, missing properties and allOf in {schema_path}"
            schema = flatten_schema(schema)
        return schema

    def get_failed_message(self, job_id):
        return (
            f"Job {job_id} has failed. To retry, please navigate to execution directory: {self._get_exec_dir(job_id)}, "
            f"get the Nextflow run command using 'nextflow log' and re-run the workflow "
            f"run command with -resume flag: 'nextflow run ... -resume'"
        )



#!/usr/bin/env python

import typer
import os
import sys
import runpy
import secrets

app = typer.Typer(pretty_exceptions_enable=False)


@app.command(name="app", context_settings={"allow_extra_args": True, "ignore_unknown_options": True})
def main(
    token: bool = typer.Option(
        False,
        "--token",
        help="Require a token to access the web app. Use set OVO_LOGIN_TOKEN env var for a custom token.",
    ),
    streamlit_run_args: typer.Context = typer.Option(None),
):
    """Run OVO streamlit web application
    \n

    OVO Config variables can be overriden with env vars in this format:
    \n
    OVO_SOME_FIELD_SUBFIELD env var will override some_field.subfield
    \n

    Additional commandline arguments will be passed to `streamlit run`, for example:
    \n
    ovo app --server.address 127.0.0.1 --server.port 5001
    \n

    See `streamlit run --help` for more info.
    """
    from ovo import config, console

    streamlit_script_path = os.path.join(os.path.dirname(os.path.dirname(__file__)), "run_app.py")
    sys.argv = ["streamlit", "run", streamlit_script_path]
    sys.argv.extend(["--browser.gatherUsageStats", "0"])  # disable streamlit usage stats
    sys.argv.extend(["--server.showEmailPrompt", "0"])  # disable streamlit email prompt (first run)
    sys.argv.extend(["--logger.enableRich", "0"])  # disable rich logging to avoid weird tracebacks
    sys.argv.extend(streamlit_run_args.args)

    if token or os.environ.get("OVO_LOGIN_TOKEN") or config.auth.always_require_token:
        token_value = os.environ.get("OVO_LOGIN_TOKEN", secrets.token_urlsafe(16))
        os.environ["OVO_LOGIN_TOKEN"] = token_value
        console.print(f"OVO Login Token: {token_value}")
    elif config.auth.admin_users and not config.auth.streamlit_auth and not config.auth.hide_admin_warning:
        console.print(
            "[yellow]Warning:[/yellow] Running with admin mode enabled but no auth configured or token set, "
            "anyone who can access the app will be able to run arbitrary commands on the server. "
            "Run with --token to generate a token, use OVO_LOGIN_TOKEN env variable to set custom a login token, "
            "or set auth.always_require_token: true in the OVO config to always generate a token. "
            "Hide this message by setting auth.hide_admin_warning: true in the OVO config."
        )

    runpy.run_module("streamlit", run_name="__main__")


if __name__ == "__main__":
    app()



import glob
import os
import argparse
import traceback

import pandas as pd
import tempfile
from typing import List, Tuple
from Bio import PDB
from Bio.PDB import PDBParser
from Bio.PDB.DSSP import DSSP
import numpy as np
import MDAnalysis as mda
from MDAnalysis.analysis import dihedrals
from MDAnalysis.analysis.data.filenames import Rama_ref


def call_dssp(pdb_path):
    structure = PDBParser().get_structure("struct", pdb_path)
    return DSSP(structure[0], pdb_path, dssp="mkdssp")


def prepare_pdb(pdb_path, pdb_path_clean, chain_ids: List[str] = None):
    """Make sure that HEADER is on first row in the PDB file to avoid breaking DSSP input
    and only keep ATOM, HETATM and LINK entries
    """
    header_idx = None
    with open(pdb_path) as f:
        for i, line in enumerate(f):
            if line.upper().startswith("HEADER"):
                header_idx = i
                if i > 0:
                    print(f"Skipping first {i} lines of PDB file before HEADER")
                break

    with open(pdb_path_clean, "wt") as w:
        if header_idx is None:
            w.write("HEADER                                                                          \n")
            header_idx = 0

        with open(pdb_path) as f:
            for i, line in enumerate(f):
                if chain_ids is not None and line.startswith("ATOM"):
                    if line[21] not in chain_ids:
                        continue
                if i >= header_idx and (
                    line.startswith("ATOM")
                    or line.startswith("HEADER")
                    or line.startswith("HETATM")
                    or line.startswith("LINK")
                ):
                    w.write(line)


def get_ramachandran_stats(angles):
    # Regions are computed from a reference set of 500 PDB files from
    # Simon C. Lovell, Ian W. Davis, W. Bryan Arendall, Paul I. W. de Bakker, J. Michael Word, Michael G. Prisant, Jane S. Richardson, and David C. Richardson.
    # Structure validation by Cα geometry: ϕ,ψ and Cβ deviation. Proteins: Structure, Function, and Bioinformatics, 50(3):437–450, January 2003. 03997.
    # URL: http://doi.wiley.com/10.1002/prot.10286 (visited on 2020-02-06), doi:10.1002/prot.10286.
    # The allowed region includes 90% data points, while the marginally allowed region includes 99% data points.
    rama_ref = np.load(Rama_ref)
    n_outliers = 0
    n_allowed = 0
    n_marginally_allowed = 0

    for phi, psi in angles:
        if pd.isna(phi) or pd.isna(psi):
            continue
        i = int((phi + 180) // 4)
        j = int((psi + 180) // 4)
        value = rama_ref[j, i]  # careful: rows=Y, cols=X

        # Source of thresholds for allowed and marginally allowed regions:
        # https://github.com/MDAnalysis/mdanalysis/blob/91146c581eee22aa6508a4d45cc271cb4d9d1509/package/MDAnalysis/analysis/dihedrals.py#L558
        if value >= 17:
            n_allowed += 1
        elif value >= 1:
            n_marginally_allowed += 1
        else:
            n_outliers += 1
    total = angles.shape[0]
    return (n_allowed, n_marginally_allowed, n_outliers, total)


def get_shape_descriptors(pdb_path) -> dict:
    u = mda.Universe(pdb_path)

    protein = u.select_atoms(f"protein")

    if len(protein.residues) == 0:
        return {}

    # Asphericity
    # (0,1) where 0 = perfect sphere
    # See: Dima, R. I., & Thirumalai, D. (2004). Asymmetry in the shapes of folded and denatured states of proteins. J Phys Chem B, 108(21), 6564-6570. doi:10.1021/jp037128y
    asphericity = protein.asphericity()

    # Shape parameter
    # S > 0 corresponds to prolate and S < 0 represents oblate
    # See: Dima, R. I., & Thirumalai, D. (2004). Asymmetry in the shapes of folded and denatured states of proteins. J Phys Chem B, 108(21), 6564-6570. doi:10.1021/jp037128y
    shape_param = protein.shape_parameter()

    # Radius of gyration
    # Rg* = R0 * N^v (Flory Law)
    # R0 = prefactor in units of spatial distance (3A)
    # v = Flory scaling exponent
    # v = 1/3 corresponds to a perfect spherical globule -> Rg* = R0 * N^(1/3)
    # v = 0.588 corresponds to a random coil -> Rg* = R0 * N^(0.588)
    # See: Dima, R. I., & Thirumalai, D. (2004). Asymmetry in the shapes of folded and denatured states of proteins. J Phys Chem B, 108(21), 6564-6570. doi:10.1021/jp037128y
    # See: Kohn, Jonathan E et al. “Random-coil behavior and the dimensions of chemically unfolded proteins.” Proceedings of the National Academy of Sciences of the United States of America vol. 101,34 (2004): 12491-6. doi:10.1073/pnas.0403643101
    rg = protein.radius_of_gyration()

    # Compare to radius of gyration of a perfect sphere
    N_residues = len(protein.residues)
    N_cube_root = N_residues ** (1 / 3)
    rg_sphere = 3 * N_cube_root
    rg_ratio = rg / rg_sphere  # < 1 tight packing, > 1 loose packing

    # Ramachandran statistics
    rama = dihedrals.Ramachandran(protein).run()
    angles = rama.results.angles[0]
    n_allowed, n_marginally_allowed, n_outliers, total = get_ramachandran_stats(angles)
    perc_allowed = n_allowed / total * 100
    perc_marginally_allowed = n_marginally_allowed / total * 100
    perc_outliers = n_outliers / total * 100

    return {
        "asphericity": asphericity,
        "shape_parameter": shape_param,
        "radius_of_gyration": rg,
        "radius_of_gyration_normalized": rg_ratio,
        "ramachandran_allowed_perc": perc_allowed,
        "ramachandran_marginally_allowed_perc": perc_marginally_allowed,
        "ramachandran_outliers_perc": perc_outliers,
    }


def prepare_chain_group_pdbs(pdb_path, out_dir: str, chains_str: str) -> Tuple[List[str], List[str]]:
    """Prepare PDB files for each chain group specified.

    :param path: Path to PDB file
    :param out_dir: Output directory to save prepared PDB files
    :param chains_str: String specifying chain groups, ";" separated groups, "," separated chains within group.
                   Example: "A,B;C" will analyze chains A and B combined as one entry, and chain C as another entry.
                   Use "FULL" to analyze all available chains as one entry.
                   Use "EACH" to analyze each chain separately.

    :return: List of chain groups (['A', 'B', 'A,B']) and corresponding list of prepared PDB file paths
    """
    with open(pdb_path) as f:
        available_chains = sorted(
            {line[21].strip() for line in f if line.startswith("ATOM") and len(line) > 21 and line[21].strip()}
        )
    print("Available chains:", available_chains)

    chain_groups = []
    for group in chains_str.split(";"):
        if group == "EACH":
            chain_groups.extend(available_chains)
        elif group == "FULL":
            chain_groups.append(",".join(available_chains))
        else:
            if missing_chains := sorted(set(group.split(",")).difference(available_chains)):
                raise ValueError(
                    f"Requested chain {missing_chains} is missing in structure: {pdb_path}, available chains: {available_chains}"
                )
            chain_groups.append(group)

    os.makedirs(out_dir, exist_ok=True)
    paths = []
    for chain_group in chain_groups:
        pdb_path_chain_subset = os.path.join(
            out_dir, f"{os.path.basename(pdb_path).removesuffix('.pdb')}.{chain_group}.pdb"
        )
        prepare_pdb(pdb_path, pdb_path_chain_subset, chain_group.split(","))
        paths.append(pdb_path_chain_subset)

    return chain_groups, paths


def get_dssp_df(paths, chains_str: str):
    """Get dataframe with one row for each PDB and chain group

    :param paths: List of paths to PDB files
    :param chains_str: List of chain IDs to analyze, ";" separated groups, "," separated chains within group.
                   Example: "A,B;C" will analyze chains A and B combined as one entry, and chain C as another entry.
                   Use "FULL" to analyze all available chains as one entry.
                   Use "EACH" to analyze each chain separately.
    :return: DataFrame with DSSP and shape descriptors
    """
    results = []
    for i, path in enumerate(paths):
        basename = os.path.basename(path).removesuffix(".pdb")
        print(f"Processing {i + 1}/{len(paths)}: {basename}")

        chain_groups, chain_pdb_paths = prepare_chain_group_pdbs(path, "prepared_pdb", chains_str)
        for chain_group, pdb_path_chain_subset in zip(chain_groups, chain_pdb_paths):
            try:
                dssp = call_dssp(pdb_path_chain_subset)
                res_num = ",".join([f"{key[1][0]}{key[1][1]}{key[1][2]}".strip() for key in dssp.keys()])
                seq = "".join([dssp[key][1] for key in dssp.keys()])
                ss = "".join([dssp[key][2] for key in dssp.keys()])
                # rsa = [dssp[key][3] for key in dssp.keys()]
                # interface = difference between exposure in isolation and exposure when all chains are present is greater than 0.1
                # interface = ''.join(['1' if dssp[key][3] - all_chains_dssp[key][3] > 0.1 else '0' for key in dssp.keys()])

                shape_descriptors = get_shape_descriptors(pdb_path_chain_subset)
                results.append(
                    {
                        "id": basename,
                        "chain": chain_group,
                        "res_num": res_num,
                        "seq": seq,
                        "ss": ss,
                        #'rsa': rsa,
                        #'interface': interface,
                        **shape_descriptors,
                    }
                )
            except Exception as e:
                print("----")
                print("Error processing:", pdb_path_chain_subset)
                traceback.print_exc()
                print("----")
                results.append(
                    {
                        "id": basename,
                        "chain": chain_group,
                        "error": str(e),
                    }
                )

    df = pd.DataFrame(results).set_index("id")
    # H: Alpha helix (4-12)
    # B: Isolated beta-bridge residue
    # E: Strand
    # G: 3-10 helix
    # I: Pi helix
    # T: Turn
    # S: Bend
    # -: None
    ss_options = ["H", "B", "E", "G", "I", "T", "S", "-"]
    ss_columns = [f"% {option}" for option in ss_options]
    if "ss" in df.columns:
        df[ss_columns] = df["ss"].apply(
            lambda x: pd.Series([None] * len(ss_options))
            if pd.isna(x) or len(x) == 0
            else pd.Series(x.count(option) / len(x) * 100 for option in ss_options)
        )
    return df


if __name__ == "__main__":
    parser = argparse.ArgumentParser()
    parser.add_argument("input_path", type=str)
    parser.add_argument("output_csv", type=str)
    parser.add_argument("--chains", type=str, required=True, help="Chain(s) to analyze, comma-separated")
    options = parser.parse_args()

    paths = sorted(glob.glob(os.path.join(options.input_path, "*.pdb")))
    if not paths:
        raise ValueError(f"No pdb files found in: {options.input_path}")

    print(f"Processing {len(paths):,} structures")
    df = get_dssp_df(paths, chains_str=options.chains)
    print(df)
    df.to_csv(options.output_csv)
    print("Saved to:", options.output_csv)



import glob
import os
import time
import argparse
from colabdesign import mk_af_model
import json

import numpy as np
from Bio import PDB
from io import StringIO


REMARK_KEYS = ["Input contig", "Standardized contig", "Chains", "Input hotspots", "Standardized hotspots"]


def get_remark_header(pdb_path: str) -> tuple[str, list[str]]:
    """
    Get the REMARK header from the PDB file.
    :param pdb_path: str, path to the PDB file
    :return: str, REMARK header
    """
    with open(pdb_path, "r") as f:
        lines = f.readlines()

    all_remark_lines = []
    remark_header = ""
    for line in lines:
        if line.startswith("REMARK   1"):
            all_remark_lines.append(line.strip().removeprefix("REMARK   1 "))
            remark_header += line

    return remark_header, all_remark_lines


def parse_remark_lines(stripped_remark_lines: list[str]) -> dict[str, str]:
    """
    :param stripped_remark_lines:
        str, lines from the PDB file header, stripped of the "REMARK   1" prefix
    :return: dict, parsed remarks: {
       "Input contig": "A45-46/10-15/A45-46/0 B24-26/5/B24-26/0 C10-20",
       "Standardized contig": "A45-46/13-13/A45-46/0 B24-26/5-5/B24-26/0 C10-20/0",
       "Chains": "A B C",
       ...
    }

    Example: example header to read
        REMARK   1 Input contig: "A45-46/10-15/A45-46/0 B24-26/5/B24-"
        REMARK   1 Input contig: "26/0 C10-20"
        REMARK   1 Standardized contig: "A45-46/13-13/A45-46/0 B24-26/5-5/B24-26/"
        REMARK   1 Standardized contig: "0 C10-20/0"
        REMARK   1 Chains: "A B C"
        REMARK   1 Input hotspots:
        REMARK   1 Standardized hotspots:
    """

    parsed_remark = {}
    for key in REMARK_KEYS:
        # Get lines that begin with this key
        remark_lines = [line.removeprefix(f"{key}:").strip() for line in stripped_remark_lines if line.startswith(key)]
        if not remark_lines and key not in ["Input hotspots", "Standardized hotspots"]:
            raise ValueError(f"Missing REMARK line for key: {key}")

        values = []
        for line in remark_lines:
            if not line and key not in ["Input hotspots", "Standardized hotspots"]:
                raise ValueError("Empty JSON value after key")
            if not line and key in ["Input hotspots", "Standardized hotspots"]:
                continue

            values.append(json.loads(line))

        parsed_remark[key] = "".join(values)

    return parsed_remark


def get_designed_residues_mask(
    remark_dict: dict[str, str],
) -> tuple[list[tuple[str, list[int]]], list[tuple[str, list[int]]], list[str]]:
    """
    Get the designed residues from the PDB file (True = designed, False = fixed)
    :Arg:
        remark_dict: dict, parsed remarks
    :Return:
        input_motif_residues: residue numbers of fixed regions in the input PDB (source motif),
                              for this input contig: 'A123-124/5/A10-11/5/B1-2
                              returns: [('A', [123, 124]), ('A', [10, 11]), ('B', [1, 2])]
        output_motif_residues: same as input_motif_residues, but with respect to the initial guess output PDB
        designed_residues: list[str], designed residues in the format [True, True, False, False, ...]
    """

    contigs_str = remark_dict.get("Standardized contig", None)
    chains_str = remark_dict.get("Chains", None)
    assert contigs_str, "No Standardized contigs found."
    assert chains_str, "No Standardized contigs found."

    contig_chains = contigs_str.split(" ")
    chains = chains_str.split(" ")

    assert len(contig_chains) == len(chains), (
        f"Number of contigs and chains do not match: {len(contig_chains)} != {len(chains)}"
    )

    input_motif_residues = []
    output_motif_residues = []
    designed_residues_mask = []
    for output_chain_id, contig_chain in zip(chains, contig_chains):
        segments = contig_chain.removesuffix("/0").split("/")
        is_fixed_chain = all(segment[0].isalpha() for segment in segments)
        standard_res = 1
        first_input_res = None
        for segment in segments:
            if not segment[0].isalpha():
                # designed segment
                start_len, end_len = map(int, segment.split("-"))
                assert start_len == end_len, (
                    f"Start len and end len of design segment should coincide: "
                    f"{start_len} != {end_len} in {contig_chain} "
                )
                designed_residues_mask.extend([True] * start_len)
                standard_res += start_len
            else:
                # fixed segment
                input_chain_id = segment[0]
                start, end = map(int, segment[1:].split("-"))
                if first_input_res is None:
                    first_input_res = start
                length = end - start + 1
                input_numbering = list(range(start, end + 1))
                standard_numbering = list(range(standard_res, standard_res + length))
                input_motif_residues.append((input_chain_id, input_numbering))

                # TODO enable this when we add standardization to initial guess output PDB
                # our standardization procedure will keep original numbering if there are no designed segments
                # so use input numbering for those chains, otherwise use consecutive numbering from 1
                #
                # output_motif_residues.append((output_chain_id, input_numbering if is_fixed_chain else standard_numbering))
                # AF2 starts numbering from 1, but then preserves the chain breaks
                af2_numbering = [r - first_input_res + 1 for r in input_numbering]
                output_motif_residues.append((output_chain_id, af2_numbering if is_fixed_chain else standard_numbering))

                designed_residues_mask.extend([False] * length)
                standard_res += length

    return input_motif_residues, output_motif_residues, designed_residues_mask


def align_multiple_proteins_pdb(
    pdb_strs: list[str], chain_residue_mappings: list[list[tuple[str, list[int] | None]] | None], all_atom: bool = False
) -> float:
    """Aligns multiple protein structures based on their atoms (CA or all).

    :param pdb_strs: list of PDB strings
    :param chain_residue_mappings: list of lists of tuples with chain ID and residues to align,
                                   if None provided, then whole chain/structure is aligned
    :param all_atom: if True, align using all atoms from matched residues (not just CA atoms)
    """
    assert len(pdb_strs) == len(chain_residue_mappings), (
        f"Expected same number of structures and chain residue mappings, got {len(pdb_strs)} != {len(chain_residue_mappings)}"
    )

    parser = PDB.PDBParser(QUIET=True)
    structures = [parser.get_structure(f"Protein{i + 1}", StringIO(pdb_str)) for i, pdb_str in enumerate(pdb_strs)]
    coords_list = []
    residues_list = []

    for i, (structure, mappings) in enumerate(zip(structures, chain_residue_mappings)):
        structure_coords = []
        structure_residues = []

        if not mappings:
            coords, res_final = get_atom_coordinates(structure, None, None, all_atom=all_atom)
            if coords:
                coords_list.append(coords)
                residues_list.append(res_final)
            else:
                raise ValueError(f"No atoms found in structure {i + 1}")

        for chain_id, residues in mappings:
            coords, res_final = get_atom_coordinates(structure, chain_id, residues, all_atom=all_atom)
            if coords:
                structure_coords.extend(coords)
                structure_residues.extend(res_final)
            else:
                raise ValueError(f"No atoms found in structure {i + 1} for chain {chain_id} and residues {residues}")

        coords_list.append(structure_coords)
        residues_list.append(structure_residues)

    for residues in residues_list:
        assert len(residues) == len(residues_list[0]), (
            f"Got different number of residues in structures: {len(residues)} != {len(residues_list[0])}"
        )

    super_imposer = PDB.Superimposer()

    for i in range(1, len(structures)):
        ref_atoms = []
        mod_atoms = []
        for atom_id, (ref_coords, mod_coords) in enumerate(zip(coords_list[0], coords_list[i])):
            shared_atom_names = sorted(set(ref_coords.keys()) & set(mod_coords.keys()))
            for atom_name in shared_atom_names:
                atom = atom_name[0]
                ref_atoms.append(PDB.Atom.Atom("X", ref_coords[atom_name], 1.0, 1.0, " ", "X", atom_id, atom))
                mod_atoms.append(PDB.Atom.Atom("X", mod_coords[atom_name], 1.0, 1.0, " ", "X", atom_id, atom))
        super_imposer.set_atoms(ref_atoms, mod_atoms)

    return super_imposer.rms


def get_atom_coordinates(
    structure: PDB.Structure.Structure,
    chain_id: str | None,
    residues: list[int] | None,
    all_atom: bool = False,
    model_index: int = 0,
) -> tuple[list[dict[str, np.ndarray]], list[PDB.Residue.Residue]]:
    coords = []
    res_final = []

    model = structure[model_index]

    for chain in model:
        if chain_id and chain.id != chain_id:
            continue
        for residue in chain:
            if residues and residue.id[1] not in residues:
                continue
            coords_dict = {}
            for atom in residue:
                if all_atom or atom.id == "CA":
                    coords_dict[atom.id] = atom.coord
            if coords_dict:
                coords.append(coords_dict)
                res_final.append(residue)
    return coords, res_final


def get_pdb_total_length(pdb_path):
    unique_residues = set()

    with open(pdb_path, "r") as f:
        for line in f:
            if line.startswith(("ATOM", "HETATM")):
                chain_id = line[21].strip()
                res_num = line[22:26].strip()
                ins_code = line[26].strip()  # column 27
                unique_residues.add((chain_id, res_num, ins_code))

    return len(unique_residues)


METRICS = {
    "rmsd": "design_backbone_rmsd",  # aligned RMSD of full structure
    "plddt": "plddt",  # 0-100
    "pae": "pae",  # predicted aligned error
    "ptm": "ptm",  # predicted TM score (0 = worst, 1 = best)
    "con": "intra_con_loss",  # intramolecular contacts loss of each residue to 2 nearest neighbors (by default) within the chain, excluding immediate sequence neighbours
    # in case of multi-chain scaffold designs
    "i_pae": "ipae",
    "i_ptm": "iptm",  # predicted interface TM score (0 = worst, 1 = best)
    "i_con": "icon_loss",  # intermolecular contacts loss of each residue to 1 nearest neighbor (by default) in another chain
}

if __name__ == "__main__":
    parser = argparse.ArgumentParser()
    parser.add_argument("input_dir", type=str)
    parser.add_argument("output_name", type=str)
    parser.add_argument("--native-pdb", required=False, type=str, help="Path to native PDB structure")
    parser.add_argument("--params", required=True, type=str, help="Path to AlphaFold2 parameter data dir")
    parser.add_argument(
        "--num-recycles", type=int, default=3, help="Number of AlphaFold2 recycles (0, 1, 2, 3, ..., default 1)"
    )
    parser.add_argument(
        "--no-templates", action="store_true", default=False, help="Do NOT use templates for fixed regions"
    )
    parser.add_argument(
        "--multimer", action="store_true", default=False, help="Use AlphaFold multimer model (default = monomer)"
    )
    options = parser.parse_args()

    model = mk_af_model(
        protocol="fixbb",
        data_dir=options.params,
        use_templates=not options.no_templates,
        use_multimer=options.multimer,
        use_initial_guess=True,
    )
    paths = sorted(glob.glob(os.path.join(options.input_dir, "*.pdb")))
    print(f"Getting info from {len(paths):,} PDBs")

    remark_lines_by_pdb_path = {}
    total_lengths = {}
    for pdb_path in paths:
        _, remark_lines = get_remark_header(pdb_path)
        if not remark_lines and (not options.no_templates or options.native_pdb):
            raise ValueError(
                f"When running with --native-pdb or with templates, "
                f'this script requires a "standardized" PDB output with the REMARK header '
                f"(as produced by the ovo rfdiffusion-backbone workflow), "
                f"no header found in: {pdb_path}"
            )
        remark_lines_by_pdb_path[pdb_path] = remark_lines
        total_lengths[pdb_path] = get_pdb_total_length(pdb_path)

    if (num_distinct_lengths := len(set(total_lengths.values()))) > 1:
        print(f"Processing {len(paths)} PDBs with {num_distinct_lengths} distinct lengths.")
        print("Each length change will cause a spike in AF2 duration!")
    else:
        print(f"Processing {len(paths)} PDBs with identical lengths. This is optimal for AF2.")

    # sort paths by total sequence length
    # to avoid spikes in duration caused by changes in input length
    paths = sorted(paths, key=lambda p: total_lengths[p])

    if options.native_pdb:
        print(os.getcwd())
        with open(options.native_pdb, "r") as pdb_f:
            native_pdb_str = pdb_f.read()
    else:
        native_pdb_str = None

    os.makedirs(options.output_name, exist_ok=True)
    with open(options.output_name.rstrip("/") + ".jsonl", "wt") as f:
        for i, path in enumerate(paths, start=1):
            basename = os.path.basename(path).removesuffix(".pdb")
            print(f"Predicting PDB {i:,}/{len(paths):,}: {basename}")
            rm_template = False  # note that when options.no_templates=True, templates are already disabled above using use_templates=False
            input_motif_residues = None
            output_motif_residues = None
            chain = None
            if not options.no_templates or native_pdb_str:
                remark_lines = remark_lines_by_pdb_path[path]
                remark_dict = parse_remark_lines(remark_lines)
                input_motif_residues, output_motif_residues, designed_residues_mask = get_designed_residues_mask(
                    remark_dict
                )
                rm_template = designed_residues_mask
                chain = ",".join(remark_dict["Chains"].split())
            if not options.no_templates:
                print(
                    f"Including template for {len(designed_residues_mask) - sum(designed_residues_mask)} fixed residues"
                )
            else:
                print("Not including template")
            start_time = time.time()
            model.prep_inputs(
                path,
                chain=chain,
                rm_template=rm_template,
            )
            model.set_seq(mode="wildtype")
            model.set_opt(num_recycles=options.num_recycles)
            model.predict(num_models=1, verbose=False)
            metrics: dict[str, float | str] = {"id": basename}
            metrics.update({new_key: model.aux["log"].get(old_key) for old_key, new_key in METRICS.items()})
            metrics["plddt"] *= 100
            metrics["pae"] = (
                metrics["pae"] * 31.0
            )  # de-normalization of https://github.com/sokrypton/ColabDesign/blob/4c0bc6d67f8f967135ecccc135a26b3bfded25e8/colabdesign/af/loss.py#L252
            if isinstance(metrics["ipae"], float):
                metrics["ipae"] = metrics["ipae"] * 31.0
            predicted_pdb_str = model.save_pdb()
            suffix = os.path.basename(options.output_name.rstrip("/"))
            with open(os.path.join(options.output_name, f"{basename}_{suffix}.pdb"), "wt") as pdb_f:
                pdb_f.write(predicted_pdb_str)
            if native_pdb_str:
                print("Predicting native motif RMSD")
                print("  Input mapping:", input_motif_residues)
                print(" Output mapping:", output_motif_residues)
                metrics["native_motif_rmsd"] = align_multiple_proteins_pdb(
                    pdb_strs=[native_pdb_str, predicted_pdb_str],
                    chain_residue_mappings=[input_motif_residues, output_motif_residues],
                    all_atom=True,
                )
            metrics["time"] = time.time() - start_time
            metrics_str = " | ".join(f"{k} = {v:.2f}" for k, v in metrics.items() if isinstance(v, float))
            print(" Prediction done in {:.1f}s | {}".format(metrics["time"], metrics_str))
            json.dump(metrics, f)
            f.write("\n")
            f.flush()



#! /usr/bin/env python

# MIT License
#
# Copyright (c) 2022 Shintaro Minami
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

import numpy as np
from einops import repeat, rearrange

CONST_Q1Q2 = 0.084
CONST_F = 332
DEFAULT_CUTOFF = -0.5
DEFAULT_MARGIN = 1.0


def _unfold(a: np.ndarray, window: int, axis: int):
    idx = np.arange(window)[:, None] + np.arange(a.shape[axis] - window + 1)[None, :]
    unfolded = np.take(a, idx, axis=axis)
    return np.moveaxis(unfolded, axis - 1, -1)


def _check_input(coord):
    org_shape = coord.shape
    assert (len(org_shape) == 3) or (len(org_shape) == 4), (
        "Shape of input tensor should be [batch, L, atom, xyz] or [L, atom, xyz]"
    )
    coord = repeat(coord, "... -> b ...", b=1) if len(org_shape) == 3 else coord
    return coord, org_shape


def _get_hydrogen_atom_position(coord: np.ndarray) -> np.ndarray:
    # A little bit lazy (but should be OK) definition of H position here.
    vec_cn = coord[:, 1:, 0] - coord[:, :-1, 2]
    vec_cn = vec_cn / np.linalg.norm(vec_cn, axis=-1, keepdims=True)
    vec_can = coord[:, 1:, 0] - coord[:, 1:, 1]
    vec_can = vec_can / np.linalg.norm(vec_can, axis=-1, keepdims=True)
    vec_nh = vec_cn + vec_can
    vec_nh = vec_nh / np.linalg.norm(vec_nh, axis=-1, keepdims=True)
    return coord[:, 1:, 0] + 1.01 * vec_nh


def get_hbond_map(
    coord: np.ndarray,
    donor_mask: np.ndarray = None,
    cutoff: float = DEFAULT_CUTOFF,
    margin: float = DEFAULT_MARGIN,
    return_e: bool = False,
) -> np.ndarray:
    # check input
    coord, org_shape = _check_input(coord)
    b, l, a, _ = coord.shape
    # add pseudo-H atom if not available
    assert (a == 4) or (a == 5), "Number of atoms should be 4 (N,CA,C,O) or 5 (N,CA,C,O,H)"
    h = coord[:, 1:, 4] if a == 5 else _get_hydrogen_atom_position(coord)
    # distance matrix
    nmap = repeat(coord[:, 1:, 0], "... m c -> ... m n c", n=l - 1)
    hmap = repeat(h, "... m c -> ... m n c", n=l - 1)
    cmap = repeat(coord[:, 0:-1, 2], "... n c -> ... m n c", m=l - 1)
    omap = repeat(coord[:, 0:-1, 3], "... n c -> ... m n c", m=l - 1)
    d_on = np.linalg.norm(omap - nmap, axis=-1)
    d_ch = np.linalg.norm(cmap - hmap, axis=-1)
    d_oh = np.linalg.norm(omap - hmap, axis=-1)
    d_cn = np.linalg.norm(cmap - nmap, axis=-1)
    # electrostatic interaction energy
    e = np.pad(CONST_Q1Q2 * (1.0 / d_on + 1.0 / d_ch - 1.0 / d_oh - 1.0 / d_cn) * CONST_F, [[0, 0], [1, 0], [0, 1]])
    if return_e:
        return e
    # mask for local pairs (i,i), (i,i+1), (i,i+2)
    local_mask = ~np.eye(l, dtype=bool)
    local_mask *= ~np.diag(np.ones(l - 1, dtype=bool), k=-1)
    local_mask *= ~np.diag(np.ones(l - 2, dtype=bool), k=-2)
    # mask for donor H absence (Proline)
    donor_mask = np.array(donor_mask).astype(float) if donor_mask is not None else np.ones(l, dtype=float)
    donor_mask = repeat(donor_mask, "l1 -> l1 l2", l2=l)
    # hydrogen bond map (continuous value extension of original definition)
    hbond_map = np.clip(cutoff - margin - e, a_min=-margin, a_max=margin)
    hbond_map = (np.sin(hbond_map / margin * np.pi / 2) + 1.0) / 2
    hbond_map = hbond_map * repeat(local_mask, "l1 l2 -> b l1 l2", b=b)
    hbond_map = hbond_map * repeat(donor_mask, "l1 l2 -> b l1 l2", b=b)
    # return h-bond map
    hbond_map = np.squeeze(hbond_map, axis=0) if len(org_shape) == 3 else hbond_map
    return hbond_map


def assign(coord: np.ndarray, donor_mask: np.ndarray = None) -> np.ndarray:
    # check input
    coord, org_shape = _check_input(coord)
    # get hydrogen bond map
    hbmap = get_hbond_map(coord, donor_mask=donor_mask)
    hbmap = rearrange(hbmap, "... l1 l2 -> ... l2 l1")  # convert into "i:C=O, j:N-H" form
    # identify turn 3, 4, 5
    turn3 = np.diagonal(hbmap, axis1=-2, axis2=-1, offset=3) > 0.0
    turn4 = np.diagonal(hbmap, axis1=-2, axis2=-1, offset=4) > 0.0
    turn5 = np.diagonal(hbmap, axis1=-2, axis2=-1, offset=5) > 0.0
    # assignment of helical sses
    h3 = np.pad(turn3[:, :-1] * turn3[:, 1:], [[0, 0], [1, 3]])
    h4 = np.pad(turn4[:, :-1] * turn4[:, 1:], [[0, 0], [1, 4]])
    h5 = np.pad(turn5[:, :-1] * turn5[:, 1:], [[0, 0], [1, 5]])
    # helix4 first
    helix4 = h4 + np.roll(h4, 1, 1) + np.roll(h4, 2, 1) + np.roll(h4, 3, 1)
    h3 = h3 * ~np.roll(helix4, -1, 1) * ~helix4  # helix4 is higher prioritized
    h5 = h5 * ~np.roll(helix4, -1, 1) * ~helix4  # helix4 is higher prioritized
    helix3 = h3 + np.roll(h3, 1, 1) + np.roll(h3, 2, 1)
    helix5 = h5 + np.roll(h5, 1, 1) + np.roll(h5, 2, 1) + np.roll(h5, 3, 1) + np.roll(h5, 4, 1)
    # identify bridge
    unfoldmap = _unfold(_unfold(hbmap, 3, -2), 3, -2) > 0.0
    unfoldmap_rev = rearrange(unfoldmap, "b l1 l2 ... -> b l2 l1 ...")
    p_bridge = (unfoldmap[:, :, :, 0, 1] * unfoldmap_rev[:, :, :, 1, 2]) + (
        unfoldmap_rev[:, :, :, 0, 1] * unfoldmap[:, :, :, 1, 2]
    )
    p_bridge = np.pad(p_bridge, [[0, 0], [1, 1], [1, 1]])
    a_bridge = (unfoldmap[:, :, :, 1, 1] * unfoldmap_rev[:, :, :, 1, 1]) + (
        unfoldmap[:, :, :, 0, 2] * unfoldmap_rev[:, :, :, 0, 2]
    )
    a_bridge = np.pad(a_bridge, [[0, 0], [1, 1], [1, 1]])
    # ladder
    ladder = (p_bridge + a_bridge).sum(-1) > 0
    # H, E, L of C3
    helix = (helix3 + helix4 + helix5) > 0
    strand = ladder
    loop = ~helix * ~strand
    onehot = np.stack([loop, helix, strand], axis=-1)
    onehot = rearrange(onehot, "1 ... -> ...") if len(org_shape) == 3 else onehot
    return onehot


atomnum = {" N  ": 0, " CA ": 1, " C  ": 2, " O  ": 3}


# Credits: Shintaro Minami, adapted from pydssp to enable filtering by chain id
def read_pdbtext_with_checking(pdbstring: str, return_sequence: bool = False, chain_id=None):
    lines = pdbstring.split("\n")
    coords, atoms, resid_old, resname_old, sequence, check = [], None, None, None, [], []
    for l in lines:
        if l.startswith("ATOM"):
            if chain_id is not None and l[21] != chain_id:
                continue
            iatom = atomnum.get(l[12:16], None)
            resid = l[21:26]
            resname = l[17:20]
            if resid != resid_old:
                if atoms is not None:
                    coords.append(atoms)
                    sequence.append(resname_old)
                    check.append(atom_check)
                atoms, resid_old, resname_old, atom_check = [], resid, resname, []
            if iatom is not None:
                xyz = [float(l[30:38]), float(l[38:46]), float(l[46:54])]
                atoms.append(xyz)
                atom_check.append(iatom)
    if atoms is not None:
        coords.append(atoms)
        sequence.append(resname_old)
        check.append(atom_check)
    coords = np.array(coords)
    sequence = np.array(sequence)
    # check
    assert len(coords.shape) == 3, "Some required atoms [N,CA,C,O] are missing in the input PDB file"
    check = np.array(check)
    assert np.all(check[:, 0] == 0), (
        "Order of PDB line may be broken. It's required to be N->CA->C->O w/o any duplicate or lack"
    )
    assert np.all(check[:, 1] == 1), (
        "Order of PDB line may be broken. It's required to be N->CA->C->O w/o any duplicate or lack"
    )
    assert np.all(check[:, 2] == 2), (
        "Order of PDB line may be broken. It's required to be N->CA->C->O w/o any duplicate or lack"
    )
    assert np.all(check[:, 3] == 3), (
        "Order of PDB line may be broken. It's required to be N->CA->C->O w/o any duplicate or lack"
    )
    # output
    if return_sequence:
        return coords, sequence
    else:
        return coords



import streamlit as st
import html
from typing import Iterable
from humanize import naturalsize
from streamlit.delta_generator import DeltaGenerator
from streamlit_js_eval import streamlit_js_eval


def heading_with_value(heading: str, secondary_text: str, level: int = 2, font_weight: int = 800):
    """Heading with additional text in grey"""
    heading = html.escape(heading)
    secondary_text = html.escape(secondary_text)
    st.html(
        f'<h{level} style="font-weight: {font_weight}">{heading}&nbsp;&nbsp;<span style="font-weight: 200; color: #888888;">{secondary_text}</span></h{level}>'
    )


def subheading_with_value(heading: str, secondary_text: str, level: int = 3, font_weight: int = 600):
    return heading_with_value(heading, secondary_text, level=level, font_weight=font_weight)


def iter_progress(iterable: Iterable, text=None, total=None, **kwargs):
    """Tqdm-like wrapper that iterates over an iterable and shows a streamlit progress bar"""
    progress_bar = st.progress(0)
    if total is None:
        total = len(iterable)
    for i, item in enumerate(iterable):
        progress_bar.progress(i / total, text=text, **kwargs)
        yield item
    progress_bar.empty()


def get_approx_screen_width():
    try:
        return max(800, streamlit_js_eval(js_expressions="window.innerWidth"))
    except Exception as e:
        print("Failed to get screen width:", e)
        return 800


def approx_max_width(max_width, center=False):
    approx_content_width = get_approx_screen_width() - 150 - 100
    st.write(approx_content_width)
    if approx_content_width < max_width:
        # full-width
        return st.columns([1])[0]
    relative_width = max_width / approx_content_width
    padding = 1 - relative_width
    if center:
        return st.columns([padding / 2, relative_width, padding / 2])[1]
    else:
        return st.columns([relative_width, padding])[0]


def wrapped_columns(n: int, wrap=4, divider=False, **kwargs) -> list[DeltaGenerator]:
    """Create n columns, wrapping them every `wrap` columns. Each row will always have `wrap` columns

    :return: flat list of n columns
    """
    assert isinstance(n, int), "n must be an integer"
    rows = n // wrap + (1 if n % wrap > 0 else 0)
    columns = []
    for row in range(rows):
        if row != 0 and divider:
            st.divider()
        for c in st.columns(wrap, **kwargs):
            if len(columns) == n:
                break
            columns.append(c)
    return columns


def confirm_download_button(data, **kwargs):
    size_suffix = f" ({naturalsize(len(data))})" if len(data) > 1024 * 1024 else ""
    st.download_button("Confirm download" + size_suffix, type="primary", data=data, **kwargs)



from typing import List, Tuple, Literal

import matplotlib.colors as mcolors
import pandas as pd

from ovo import config
from ovo.core.database import Descriptor, NumericDescriptor
from ovo.core.database.models_proteinqc import ProteinQCTool
from ovo.core.scheduler.base_scheduler import Scheduler


def tool_supports_scheduler(tool: ProteinQCTool, scheduler: Scheduler) -> bool:
    if "profile" in scheduler.submission_args and "conda" in scheduler.submission_args["profile"]:
        return tool.supports_conda
    return True


def get_available_schedulers(tools: List[ProteinQCTool]) -> dict[str, Scheduler]:
    """
    Get available schedulers based on the tools selected by the user.
    """
    available_schedulers = {}

    for scheduler_key, scheduler in config.schedulers.items():
        if all(tool_supports_scheduler(tool, scheduler) for tool in tools):
            available_schedulers[scheduler_key] = scheduler
    return available_schedulers


def get_available_tools(tools: List[ProteinQCTool], scheduler: Scheduler) -> List[ProteinQCTool]:
    """
    Get available tools based on the selected scheduler.
    """
    return [tool for tool in tools if tool_supports_scheduler(tool, scheduler)]


def get_descriptor_plot_setting(descriptor: Descriptor) -> Tuple[List[float], bool, str]:
    """
    Get the thresholds, reverse_colors and color_by settings for the descriptor plot.

    The logic for thresholds is as follows:
    - For descriptors with 'lower_is_better', thresholds are [min_value, warning_value, error_value, max_value].
    - For descriptors with 'higher_is_better', thresholds are [min_value, error_value, warning_value, max_value].
    - If any of these values are not set (i.e., None), thresholds will be None.

    Returns:
        Tuple[List[float], bool, str]: thresholds, reverse_colors, color_by
    """
    reverse_colors = False
    if descriptor.comparison == "lower_is_better":
        thresholds = [descriptor.min_value, descriptor.warning_value, descriptor.error_value, descriptor.max_value]
        if None in thresholds:
            thresholds = None
            color_by = "lower_is_better"
        else:
            reverse_colors = True
            color_by = "thresholds"
    elif descriptor.comparison == "higher_is_better":
        thresholds = [descriptor.min_value, descriptor.error_value, descriptor.warning_value, descriptor.max_value]
        if None in thresholds:
            thresholds = None
            color_by = "higher_is_better"
        else:
            color_by = "thresholds"
    else:
        thresholds = None
        reverse_colors = False
        color_by = None

    if isinstance(descriptor, NumericDescriptor) and descriptor.color_scale:
        color_by = descriptor.color_scale

    return thresholds, reverse_colors, color_by


def get_thresholds_colormap(thresholds: List[float], reverse_colors: bool = False, colors=None):
    """
    Create a matplotlib colormap and normalization object corresponding to the Altair redyellowgreen scheme with 4 points in the domain as specified by the thresholds.

    Args:
        thresholds (List[float]): List of threshold values to define color transitions.
        reverse_colors (bool): If True, reverse the color order.

    Returns:
        Callable[[float], str]: A function that takes a value and returns a hex color string of a color in the colormap.
    """

    # Altair redyellowgreen scheme with 4 points in the domain
    assert len(thresholds) == 4, "Expected 4 thresholds for colormap"
    assert None not in thresholds, "Thresholds must not contain None values"

    if colors is None:
        colors = [
            "#f16e43",
            "#fedd8d",
            "#d7ee8e",
            "#64bc61",
        ]  # https://vega.github.io/vega/docs/schemes/#redyellowgreen

    if reverse_colors:
        colors = colors[::-1]

    # Normalize thresholds to [0, 1] for colormap construction
    t_min, t_max = min(thresholds), max(thresholds)
    normalized_positions = [(t - t_min) / (t_max - t_min) for t in thresholds]

    cmap = mcolors.LinearSegmentedColormap.from_list("threshold_colormap", list(zip(normalized_positions, colors)))

    # Use Normalize to map the values onto the domain
    norm = mcolors.Normalize(vmin=t_min, vmax=t_max)

    return lambda value: mcolors.to_hex(cmap(norm(value)))


def get_plddt_color(thresholds: List[float] = [90, 70, 50], lighter: bool = False):
    """
    Create a color function for pLDDT values based on standard thresholds.

    # > 90, very high confidence, '#0053d6'
    # > 70, high confidence, '#65cbf3'
    # > 50, low confidence, '#ffdb13'
    # < 50, 'Very low confidence', '#ff7d45'
    """

    colors = ["#0053d6", "#65cbf3", "#ffdb13", "#ff7d45"]
    if lighter:
        colors = ["#2a6ed1", "#7fd0f8", "#ffd24d", "#ff8f5e"]
    assert len(colors) == len(thresholds) + 1, "Number of colors must be one more than number of thresholds"

    def color_func(value):
        for t, color in zip(thresholds, colors):
            if value >= t:
                return color
        return colors[-1]

    return color_func


def get_pae_colormap():
    """
    Create a color function for PAE values.
    """
    return get_neutral_colormap(0, 25)


def get_rmsd_colormap():
    """
    Create a color function for RMSD values.
    """
    return get_thresholds_colormap([0, 2, 6, 13], reverse_colors=True)


def get_higher_is_better_colormap(min_val, max_val):
    colors = ["#ffffff", "#d7ee8e"]  # white → light green
    cmap = mcolors.LinearSegmentedColormap.from_list("neutral_accent", colors)

    def color_func(value):
        if pd.isna(value):
            return ""
        # Normalize value to [0, 1] based on min/max
        normed = (value - min_val) / (max_val - min_val) if max_val != min_val else 0.0
        return mcolors.to_hex(cmap(normed))

    return color_func


def get_lower_is_better_colormap(min_val, max_val):
    colors = ["#d7ee8e", "#ffffff"]  # light green -> white
    cmap = mcolors.LinearSegmentedColormap.from_list("neutral_accent", colors)

    def color_func(value):
        if pd.isna(value):
            return ""
        # Normalize value to [0, 1] based on min/max
        normed = (value - min_val) / (max_val - min_val) if max_val != min_val else 1.0
        return mcolors.to_hex(cmap(normed))

    return color_func


def get_neutral_colormap(min_val, max_val):
    colors = ["#ffffff", "#b0b0b0"]  # white -> light gray
    cmap = mcolors.LinearSegmentedColormap.from_list("neutral_accent", colors)

    def color_func(value):
        if pd.isna(value):
            return ""
        # Normalize value to [0, 1] based on min/max
        normed = (value - min_val) / (max_val - min_val) if max_val != min_val else 0
        return mcolors.to_hex(cmap(normed))

    return color_func


def get_descriptor_cmap(descriptor: Descriptor, min_val: float, max_val: float):
    """
    Return a color mapping function for the given descriptor.

    Colormap logic:
    - For special descriptors (e.g., pLDDT, PAE), apply custom color schemes.
    - If all thresholds (min_value, warning_value, error_value, max_value) are set, use them for color transitions based on the comparison type.
    - If thresholds are not fully set but a comparison is defined, use a 2-color scale (more green is better) based on comparison type and min/max values in the data.
    - If thresholds are not fully set and no comparison is defined, use a neutral colormap (higher values are darker grey), determined by min/max values in the data.

    Args:
        descriptor (Descriptor): The descriptor for which to create the colormap.
        min_val (float): Minimum value among the selected designs.
        max_val (float): Maximum value among the selected designs.
    """
    if not isinstance(descriptor, NumericDescriptor):
        return lambda val: "#ffffff"

    thresholds, reverse_colors, color_by = get_descriptor_plot_setting(descriptor)

    if descriptor.min_value is not None and descriptor.max_value is not None:
        # When both min_value and max_value are explicitly set in the descriptor, use them instead of data min/max
        min_val, max_val = descriptor.min_value, descriptor.max_value

    return get_cmap(
        color_by=color_by,
        min_val=min_val,
        max_val=max_val,
        thresholds=thresholds,
        reverse_colors=reverse_colors,
    )


def get_cmap(color_by, min_val: float, max_val: float, thresholds: list[float] = None, reverse_colors: bool = False):
    if pd.isna(min_val) or pd.isna(max_val):
        return lambda val: "#ffffff"

    if color_by == "plddt":
        cmap = get_plddt_color(lighter=True)
    elif color_by == "pae":
        cmap = get_pae_colormap()
    elif color_by == "rmsd":
        cmap = get_rmsd_colormap()
    elif color_by == "thresholds" and thresholds is not None:
        cmap = get_thresholds_colormap(thresholds, reverse_colors=reverse_colors)
    elif color_by == "higher_is_better":
        cmap = get_higher_is_better_colormap(min_val, max_val)
    elif color_by == "lower_is_better":
        cmap = get_lower_is_better_colormap(min_val, max_val)
    else:
        cmap = get_neutral_colormap(min_val, max_val)

    return cmap


def get_flag_color(value: float, descriptor: Descriptor) -> Literal["green", "yellow", "orange", "missing", None]:
    if not isinstance(descriptor, NumericDescriptor):
        return None
    if not descriptor.warning_value or not descriptor.error_value:
        return None
    if pd.isna(value):
        return "missing"
    try:
        value = float(value)
    except ValueError:
        return None
    if descriptor.comparison == "higher_is_better":
        if value > descriptor.warning_value:
            return "green"
        elif value > descriptor.error_value:
            return "yellow"
        else:
            return "orange"
    elif descriptor.comparison == "lower_is_better":
        if value < descriptor.warning_value:
            return "green"
        elif value < descriptor.error_value:
            return "yellow"
        else:
            return "orange"
    else:
        return None


def get_descriptor_comment(descriptor: Descriptor) -> str:
    if not isinstance(descriptor, NumericDescriptor):
        return descriptor.description

    thresholds, _, color_by = get_descriptor_plot_setting(descriptor)

    description = descriptor.description

    if color_by == "plddt":
        description += f"\n(pLDDT color scheme: > 90 very high confidence, > 70 high confidence, > 50 low confidence, < 50 very low confidence)"
    elif color_by == "pae":
        description += f"\n(PAE color scheme: lower values are better, higher values indicate less confidence)"
    elif thresholds is not None and color_by == "thresholds":
        description += f"\n(Minimum value: {descriptor.min_value}, Maximum value: {descriptor.max_value},  Warning value: {descriptor.warning_value}, Error value: {descriptor.error_value})"

    return description



#!/usr/bin/env python3
#
# Calculates contact molecular surface and various other metrics suggested by
# Brian Coventry. XML is unaltered from Brian.
#
# Adapted by David Prihoda for RFpeptide scoring based on RFpeptide supplement
#
# Usage:
#
#   ./pyrosetta_interface_metrics.py PDB_FOLDER OUT_CSV
#
# where FOLDER contains PDBs of complexes you want to score.
#


import os
import glob
import argparse
import multiprocessing
import json
import sys

from pyrosetta import *
from rosetta.protocols.rosetta_scripts import *

script_dir = os.path.dirname(__file__)

p = argparse.ArgumentParser()
p.add_argument("data_dir", help="Folder of FastDesign outputs to process. Must contain both binder and receptor.")
p.add_argument("out_jsonl", help="Output jsonl file name")
p.add_argument(
    "--suffix",
    default="",
    help="Suffix on outputs to process. e.g. if --suffix=complex, then only processes files named *complex.pdb",
)
p.add_argument("--relax", action="store_true", default=False, help="Relax structures before scoring")
p.add_argument("--out-pdb", help="Save PDB structures to this directory after applying movers")
p.add_argument("--debug", action="store_true", default=False, help="Exit on error")
args = p.parse_args()

# TODO include "-holes:dalphaball /software/rosetta/DAlphaBall.gcc"
# As in https://github.com/RosettaCommons/RFDesign/blob/main/scripts/get_interface_metrics.py
# To enable BuriedUnsatHbonds in interface_metrics_rfpeptides.xml
init("-corrections::beta_nov16 -detect_disulf false -run:preserve_header true")
parser = RosettaScriptsParser()
protocol_path = script_dir + "/interface_metrics_rfdesign.xml"

ncpu = len(os.sched_getaffinity(0)) if hasattr(os, "sched_getaffinity") else os.cpu_count()
print(f"Using {ncpu} cores")


def calculate(pdb_path):
    basename = os.path.basename(pdb_path).removesuffix(".pdb")
    row = {"id": basename}
    try:
        pose = pose_from_pdb(pdb_path)

        if args.relax:
            objs = XmlObjects.create_from_file(protocol_path)
            relax = objs.get_mover("FastRelax")
            relax.apply(pose)

        # We re-initialize the mover for each PDB to avoid remembering stuff from previous peptides, like peptide length
        protocol = parser.generate_mover(protocol_path)
        # Apply movers
        protocol.apply(pose)

        if args.out_pdb:
            os.makedirs(args.out_pdb, exist_ok=True)
            # save pose
            pose.dump_pdb(os.path.join(args.out_pdb, str(basename) + "_relaxed.pdb"))

        print("SCORES", pose.scores)
        for k, v in pose.scores.items():
            row[k] = float(v)
    except Exception as e:
        row["error"] = f"{e} ({type(e).__name__})"
        print(f"ERROR processing {basename}: {row['error']}")
        if args.debug:
            raise
    return row


if __name__ == "__main__":
    files = sorted(glob.glob(os.path.join(args.data_dir, f"*{args.suffix}.pdb")))

    if not files:
        print("No files found in", args.data_dir)
        sys.exit(0)

    if args.debug:
        records = map(calculate, files)
    else:
        with multiprocessing.Pool(min(ncpu, len(files))) as p:
            records = p.map(calculate, files)

    num_errors = 0
    with open(args.out_jsonl, "wt") as f:
        for record in records:
            json.dump(record, f)
            if record.get("error"):
                num_errors += 1
            f.write("\n")

    if num_errors == len(files):
        raise ValueError(f"All results failed: {record['error']}")

    print("Saved to:", args.out_jsonl)
